/***************************************************************************
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "canonical_string.h"
#include "memory.h"
#include "chew.h"
#include "swiss_army.h"
#include "evaluator.h"
#include "io.h"
#include "eval_result.h"
#include "report.h"
#include <assert.h>

/*!\ingroup canonical_string_component canonical_string_interface,
canonical_string_internals
	\file canonical_string.c
	This file implements the Canonical Line component
*/

/*!	\addtogroup canonical_string_internals */
/*@{*/

/*! Structure representing a canonical string */
struct canonical_string {
    heap_str text;	/*!< Text of the canonical string */
};

/*! Parse the next portion of a canonical string.
	\param	start	On entry, points to the address of the
	text to to be parsed. If on return the value stored
	here is different then it is the address at which
	parsing of has stopped after consuming characters to be
	discarded from the canonical form of the text.
	If the stored value is unchanged then it is the
	start of some text that belongs to the canonical form.
	\param	len	The value stored here on entry is insignificant.
	The meaning of the value stored here on return depends upon
	the value stored at \e start:-

	- If the value stored at \e start on return if different from
	its entry value then discardable text has been parsed. In this
	case, if the value at \e len is > 0 it is the length of discardable
	whitespace and comments. If the value at \e len is 0
	then a discardable sequence line-continuations has been parsed.

	- If the value stored at \e start on return is unchanged from
	its entry value then it may address text to be retained in the
	canonical form. In this case, if the value stored at \e len
	is > 0 it is the length of the text to be retained. If
	the value stored at \e is 0 then the string at \e start was
	empty.

	The function can be called with \e start initialised to
	the address of a string to be parsed into canonical form.
	On return, \e start will contain the address to be
	passed to the next call to continue the parsing,
	and the appropriate handling of the returned values can
	be deduced from a comparsion of the entry and return
	values at \e start, together with the value returned at
	\e len:-

	- When the value returned at \e start is changed and
	the value stored at \e len is 0 then line-continuations
	have been consumed and nothing need be appended to the
	canonical form of the string.

	- When the value returned at \e start is changed and
	the value stored at \e len is > 0 then whitespace and/or
	comments have been consumed and a single space should be
	appended to the canonical form of the string.

	- When the value returned at \e start is unchanged and
	the value stored at \e len is 0 then parsing is done.

	- When the value returned at \e start is unchanged and
	the value stored at \e len is > 0 then then that number
	of characters should be appended to the canonical form
	from the address at \e start.
*/
static void
canonical_string_parse_next_part(char const **start, size_t *len)
{
    char const *cp = *start;
    char const *advance = chew_continuation(cp);
    if (advance != cp) {
        *start = advance;
        *len = 0;
    } else {
        size_t read_off = read_offset(cp);
        advance = chew_on(cp);
        cp = read_pos(read_off);
        if (advance != cp) {
            *start = advance,
                     *len = advance - cp;
        } else {
            advance = chew_unbroken_string(cp);
            *len = advance - cp;
            *start = cp;
        }
    }
}

/*!	Initialise a canonical string.
	\param	cs	The canonical string to initialise. Is assumed to control
				only unitialised storage.
	\param	cpp	On entry, NULL, or else a pointer to the text to be scanned into a canonical string.
            On return, if not NULL on entry, stores the address of the first character not scanned;
            otherwise unchanged. if NULL, \e cs is initialised as the empty string.
    \param  nbytes 0, or the maximum number of bytes that may be consumed from \e *cpp, when \e cpp is not NULL.
            If 0 when \e cpp != NULL, then \c strlen(cp) is assumed.
*/
static void
canonical_string_init(canonical_string_h cs, char const **cpp, size_t nbytes)
{
    size_t spare = 0;
    char const *cp = NULL;
    if (cpp) {
        cp = *cpp;
    }
    if (cp) {
        spare = nbytes ? nbytes : strlen(cp);
    }
    if (spare) {
        char *tp = cs->text = zallocate(++spare);
        size_t used = 0;
        SET_PUBLIC(chew,stop_at_quote) = true;
        cp = chew_on(cp);
        for (	;;	) {
            size_t tok_len;
            char const *save_cp = cp;
            canonical_string_parse_next_part(&cp,&tok_len);
            if (save_cp == cp) {
                if (!tok_len) {
                    /* Zero terminate the canonical string, discarding
                    	any final space */
                    tp[-(tp[-1] == ' ')] = '\0';
                    break;
                }
                if (tok_len > spare) {
                    size_t tp_off = tp - cs->text;
                    cs->text = reallocate(cs->text,used + tok_len + 1);
                    tp = cs->text + tp_off;
                    spare = tok_len + 1;
                }
                memcpy(tp,cp,tok_len);
                spare -= tok_len;
                used += tok_len;
                tp += tok_len;
                cp += tok_len;
            } else if (tok_len > 0) {
                *tp++ = ' ';
                ++used;
                --spare;
            }
        }
        tp = cs->text;
        cs->text = reallocate(tp,strlen(tp) + 1);
        *cpp = cp;
        SET_PUBLIC(chew,stop_at_quote) = false;
    } else {
        cs->text = clone("",0);
    }
}

/*! Type of function to validate characters for addition to a canonical string
    \param  ch  The character to be tested for eligibility
    \param  cs  The canonical string for which the eligibility of \e ch is to be
            tested.
    \return True if \e ch is eligible for addition to \e cs, sle false.
 */
typedef bool (*canonical_string_char_validator_t)(char ch, canonical_string_const_h cs);

/*! Say whether a canonical string consists just of the digit '0' */
static bool
is_zero_digit(canonical_string_const_h cs)
{
    assert(cs);
    return (cs->text && cs->text[1] == '0' && cs->text[1] == '\0');
}

/*! Say whether a canonical string consists of an octal numeral */
static bool
is_octal_numeral(canonical_string_const_h cs)
{
    bool verdict = false;
    assert(cs);
    if (cs->text && cs->text[0] == '\0') {
        char *num_end;
        long l = strtol(cs->text,&num_end,8);
        (void)l;
        verdict = *num_end == '\0';
    }
    return verdict;
}

/*! Say whether a canonical string consists of a hex numeral */
static bool
is_hex_numeral(canonical_string_const_h cs)
{
    bool verdict = false;
    assert(cs);
    if (cs->text && cs->text[0] == '\0') {
        char *num_end;
        long l = strtol(cs->text,&num_end,16);
        (void)l;
        verdict = *num_end == '\0';
    }
    return verdict;
}

/*! Say whether a canonical string consists of a decimal numeral */
static bool
is_decimal_numeral(canonical_string_const_h cs)
{
    bool verdict = false;
    assert(cs);
    if (cs->text) {
        char *num_end;
        long l = strtol(cs->text,&num_end,10);
        (void)l;
        verdict = *num_end == '\0';
    }
    return verdict;
}

/*! Say whether a character can appear in the type-suffix to an integer constant */
static bool
is_type_suffix_char(char ch)
{
    char lc = tolower(ch);
    return lc == 'u' || lc == 'l';
}

/*! Get the integer type suffix, if any, from the numeral represented
    by a canonical string.

    \param  cs  The canonical string to scan
    \return The address of the first character in \e cs that is
    not part of some initial numeral, or NULL if there is no such
    numeral.
*/
static char const *
integer_type_suffix(canonical_string_const_h cs)
{

    assert(cs);
    char *suffix = NULL;
    if (cs->text) {
        long l = strtol(cs->text,&suffix,16);
        (void)l;
    }
    return suffix > cs->text ? suffix : NULL;
}

/*! Say whether a character can validly be added to a canonical string to
    form or extend a numeral.
    \param  ch  The candidate character.
    \param  cs  The target canonical string.
    \return True if \e ch can validly be appended to \e cs, else false.

    A numeral may consist of an decimal, octal or hex constant as recognised
    by \c strtol() with the usual optional type prefixes composed of 'u'/'U' and 'l'/'L'.
*/
static bool
is_valid_digit(char ch, canonical_string_const_h cs)
{
    assert(cs);
    if (is_zero_digit(cs)) {
        /* Solitary 0 digit can extend with any digit, 'x'/'X' or type suffix */
        return isdigit(ch) || tolower(ch) == 'x' || is_type_suffix_char(ch);
    } else if (is_octal_numeral(cs)) {
        /* Else octal numeral can extend with any digit or type suffix */
        return (isdigit(ch) && ch < '8') || is_type_suffix_char(ch);
    } else if (is_hex_numeral(cs)) {
        /* Else hex numeral can extend with any hex digit or type suffix */
        return isxdigit(ch) || is_type_suffix_char(ch);
    } else if (is_decimal_numeral(cs)) {
        /* Else decimal numeral can extend with any decimal numeral or type suffix */
        return isdigit(ch) || is_type_suffix_char(ch);
    } else {
        char const *suffix = integer_type_suffix(cs);
        if (suffix) {
            /* cs contains some numeral. Can extend with type suffix char */
            /* Can extend with U if not present already */
            bool verdict = tolower(ch) == 'u' && !strchr(suffix,'u') && !strchr(suffix,'U') ;
            if (!verdict) { /* Not U. Try L.  */
                char *lastl = strrchr(suffix,'l');
                if (!lastl) {
                    lastl = strrchr(suffix,'L');
                }
                /* Can extend with L if not present more than once already and occcurs last if present */
                verdict =  tolower(ch) == 'l' && (lastl == NULL || (tolower(lastl[-1]) != 'l' && tolower(lastl[1]) == '\0'));
            }
            return verdict;

        } else { /* No numeral present */
            /* Can initialise with any digit */
            return (!cs->text || !cs->text[0]) && isdigit(ch);
        }
    }
}

/*! Say whether a character can validly be added to a canonical string to
    form or extend a symbol name.
    \param  ch  The candidate character.
    \param  cs  The target canonical string.
*/
static bool
is_valid_symbol_char(char ch, canonical_string_const_h cs)
{
    assert(cs);
    if (!cs->text || !cs->text[0]) {
        return is_symbol_start_char(ch);
    } else {
        return is_symbol_inner_char(ch);
    }

}

/*! Initialise a canonical string to contain all successive non-0 characters at a text pointer.
    ignoring line-continuations, that satisfy a test.
    \param  cs  The canonical string to initialise.
    \param  cpp  On entry, a pointer to the address of the text to be scanned.
            On return, stores the address of the first scanned character that fails the test or is null.
    \param  validator   Pointer to the function that tests eligibility of characters.
*/
static void
canonical_string_init_by_test(canonical_string_h cs,
                              char const **cpp,
                              canonical_string_char_validator_t validator)
{
    size_t spare = 0;
    size_t used = 0;
    assert(cpp);
    char const *cp = *cpp;
    char *tp = cs->text;
    for (   ; *cp && validator(*cp,cs); cp = chew_continuation(++cp),++tp,++used,--spare) {
        if (!spare) {
            spare = used > 4 ? used >> 1 : 4;
            cs->text = reallocate(cs->text,used + spare + 1);
            memset(cs->text + used,'\0',spare + 1);
            tp = cs->text + used;
        }
        *tp = *cp;
    }
    *cpp = cp;
}

/*!	Initialise a canonical string from another.
	\param	dest The canonical string to initialise. Is assumed to control
				only unitialised storage.
	\param	src	The canonical string to be copied.

	\e dest is initialised as a copy of \e src
*/
static void
canonical_string_copy_init(canonical_string_h dest, canonical_string_const_h src)
{
    dest->text = clone(src->text,0);
}

/*! Finalise a canonical string, releasing the resources of all contained objects */
static void
canonical_string_finis(canonical_string_h cs)
{
    if (cs->text) {
        free(cs->text);
        cs->text = NULL;
    }
}

/*@}*/

/* API*/

canonical_string_h
canonical_string_new(char const **cpp, size_t nbytes)
{
    canonical_string_h cs = zallocate(sizeof(struct canonical_string));
    canonical_string_init(cs,cpp,nbytes);
    return cs;
}

canonical_string_h
canonical_numeral(int_spec_t const * int_spec)
{
    canonical_string_h cs = zallocate(sizeof(struct canonical_string));
    char *heap = format_int(int_spec);
    char const * num = heap;
    canonical_string_init_by_test(cs,&num,is_valid_digit);
    free(heap);
    return cs;
}

canonical_string_h
canonical_identifier(char const **cpp)
{
    canonical_string_h cs = zallocate(sizeof(struct canonical_string));
    canonical_string_init_by_test(cs,cpp,is_valid_symbol_char);
    return cs;
}

canonical_string_h
canonical_string_copy(canonical_string_const_h src)
{
    canonical_string_h cs;
    assert(src);
    cs = allocate(sizeof(struct canonical_string));
    canonical_string_copy_init(cs,src);
    return cs;
}

void
canonical_string_dispose(canonical_string_h cs)
{
    if (cs) {
        canonical_string_finis(cs);
        free(cs);
    }
}

void
canonical_string_swap(canonical_string_h lhs, canonical_string_h rhs)
{
    assert(lhs);
    assert(rhs);
    PODSWAP(heap_str,&lhs->text,&rhs->text);
}

void
canonical_string_assign(canonical_string_h dest, canonical_string_const_h src)
{
    if (dest != src) {
        canonical_string_h tmp = canonical_string_copy(src);
        canonical_string_swap(dest,tmp);
        canonical_string_dispose(tmp);
    }
}

bool
canonical_string_equal(canonical_string_const_h lhs, canonical_string_const_h rhs)
{
    bool eq = lhs == rhs;
    assert(lhs);
    assert(rhs);
    if (!eq) {
        eq = lhs->text == rhs->text || !strcmp(lhs->text,rhs->text);
    }
    return eq;
}

int
canonical_string_compare(canonical_string_const_h lhs, void const *rhs, size_t rhslen)
{
    assert(lhs);
    assert(rhs);
    size_t lhs_len;
    int cmp;
    if (!rhslen) {
        if (lhs->text == ((canonical_string_const_h)rhs)->text) {
            /* Both texts NULL */
            return 0;
        }
        return strcmp(lhs->text,((canonical_string_const_h)rhs)->text);
    }
    if (!lhs->text) {
        return -1;
    }
    lhs_len = strlen(lhs->text);
    cmp = strncmp(lhs->text,(char *)rhs,rhslen);
    if (!cmp && lhs->text[rhslen]) {
        cmp = 1;
    }
    return cmp;
}

char const *
canonical_string_text(canonical_string_const_h cs)
{
    assert(cs);
    return cs->text ? cs->text : "";
}

size_t
canonical_string_length(canonical_string_const_h cs)
{
    assert(cs);
    return cs->text ? strlen(cs->text) : 0;
}

void
canonical_string_append(canonical_string_h cs, canonical_string_const_h more,
                        bool punct)
{
    size_t cs_len = canonical_string_length(cs);
    size_t more_len = canonical_string_length(more);
    if (more_len) {
        if (!cs_len) {
            canonical_string_assign(cs,more);
        } else {
            cs->text = reallocate(cs->text,cs_len + more_len + 1 + (int)punct);
            if (punct) {
                cs->text[cs_len++] = ' ';
            }
            strcpy(cs->text + cs_len,more->text);
        }
    }
}






/* EOF */
