#ifndef GET_OPTIONS_H
#define GET_OPTIONS_H
/***************************************************************************
 *   Copyright (C) 2004 - 2006 Symbian Software Ltd.                       *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "bool.h"

/*!\ingroup get_options_component get_options_interface.
 *\file get_options.h
 * This file provides the Get Options component interface
*/

/*!	\addtogroup get_options_interface*/
/*@{*/

extern int optind;
/*!< Index of option parsed by getopt_long()*/
extern int optopt;
/*!< Unrecogised short option parsed by getopt_long()*/
extern char *optarg;
/*!< Argument to an option parsed by getopt_long()*/

/*! Info structure for an option for getopt_long().*/
struct option {
    /*! The name of the option*/
    const char *name;
    /*! Can the option take an argument? One of:
    	- \c no_argument
    	- \c required_argument
    	- \c optional argument
    */
    int has_arg;
    /*! \c NULL here makes getopt_long() return the field \c val if
        this option is matched. Non-null makes getopt_long() return
        0 if any option is matched, and set the \c int addressed by
        <tt>flag == val</tt>/ if this option is matched, otherwise
        leaving it unchanged.
    */
    int *flag;
    /*!	Value to be returned or stored by getopt_long() if this
        option is matched (depending on \c flag).
    */
    int val;
};

/*! Possible values of the \c has_arg field in an \c option structure*/
enum {
    no_argument, /*!< The option cannot have an argument*/
    required_argument, /*!< The option must have an argument*/
    optional_argument /*!< The option mat or may not have an argument*/
};


/*! Info structure for a command option for get_command()*/
typedef struct cmd_option {
    /*! The name of the option*/
    const char *name;
    /*!	Value to be returned if the command is matched*/
    int cmd_code;
    /*!	Bits representing the command in a bit mask */
    int cmd_mask;
} cmd_option_t;

/*! Structure representing a set of options that are excluded
	by another option.
*/
struct exclusion_list {
    int excluder; /*!< This option code excludes the others.*/
    int const * excluded; /*!< This 0-terminated list of option codes
					is excluded by \e excluder.*/
};

/*! Look for a command option at the start of commandline arguments.
	\param	argc	The number of commandline arguments.
	\param	argv	The array of commandline arguments.
	\param	commands	An array of valid command options.
	\return If \e argc < 2 return NULL. Else if \e argv[1] matches the \e name
		field in any of	the command options listed in \e commands, return
		the address of that command option, else return NULL.
*/
extern cmd_option_t const *
get_command_option(int argc, char *argv[], cmd_option_t const * commands);

/*! Say whether options are compatible with respect to specified
	incompatbilities.
	\param	opt_excluder	To be determined if this option
			excludes another.
	\param	opt_excluded	To be determined if this option
			is excluded by \e opt_excluder.
	\param	exclusions	Addresses an array of exclusions lists,terminated
			by an exclusion list in which the \e excluder field is 0,
			in which a list belonging to \e opt_excluder will be sought.
	\param	indexed  If true then the function may assume that \e opt_excluder
			is an index to its exclusion list in \e exclusions.
	\return	False if \e exclusions is found to contain an exclusion list
			whose \e excluder field is \e opt_excluder and whose
			\e excluded field is an array containing \e opt_excluded.
			Otherwise true.

*/
extern bool
opts_are_compatible(	int opt_excluder, int opt_excluded,
                        struct exclusion_list const *exclusions, bool indexed);

/*! We need the standard getopt_long() function, but the host implementation
	cannot be relied on to be re-entrant. Therefore we roll our own
	re-entrant version. The interface is like that of its standard namesake.
*/
extern int
getopt_long(int argc, char * argv[], const char *optstr,
            struct option const *longopts, int *longind);

/*! Look up the long name of the an option in an array of \c struct \e option
	\param	longopts	The array of option info to be searched, terminated
				by a \c struct \e option in which the \e name field is NULL.
	\param	opt	The option code to be sought.
	\return The address of the long name of the option with code \e opt, if
		found, else NULL.

*/
extern const char *
get_long_opt_name(struct option const * longopts, int opt);

/*@}*/

#endif /* EOF*/

