/***************************************************************************
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the arg of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include "hash_include.h"
#include "eval_result.h"
#include "includes_tally.h"
#include "configured_symbols.h"
#include "unconfigured_symbols.h"
#include "chew.h"
#include "line_select.h"
#include "args.h"
#include "directives_tally.h"


/*!\ingroup hash_include_component hash_include_interface
	\file hash_include.c
    This file implements the Include-Argument component
*/

/*! \addtogroup hash_include_internals */
/*@{*/

/*!	Parse the resolved argument of a hash_include directive.
	\param	cp	Start of the resolved argument to be parsed.
	\param	pdiagnostic  Pointer at which will be stored on return either
        NULL or the address of a diagnostic. If NULL is stored then
        the argument of the hash-include directive is valid; otherwise
        the stored diagnostic says why it is problematic.
	\return	The address of the first byte >= cp that is not parsed.

	The function assumes that the argument at \c cp is not a symbol;
	it may be the resolved value of a symbol.
*/
static char const *
hash_include_parse_final(char const *cp, char const **pdiagnostic)
{
    char const *arg_start;
    char const *arg_end;
    size_t arglen;
    assert(cp);
    *pdiagnostic = NULL;
    arg_start = cp;
    cp = arg_end = chew_header_name(arg_start);
    arglen = arg_end - arg_start;
    if (!arglen) {
        if (!*arg_end) {
            *pdiagnostic = "#include has no argument";
        } else {
            *pdiagnostic = "bad header name";
        }
    } else if (*arg_start == '\"' || *arg_start == '<') { /* Looks like a header name */
        if (*arg_start == '\"') { /* Looks like local header */
            if (arg_end[-1] != '\"' || arglen < 3) {
                *pdiagnostic = "bad local header name";
            }
        } else if (*arg_start == '<') { /* Looks like system header */
            if (arg_end[-1] != '>' || arglen < 3) {
                *pdiagnostic = "bad system header name";
            }
        }
    }
    cp = chew_string(arg_end);
    return cp;
}
/*@}*/

/* API */

hash_include_h
hash_include_new(char const *arg, size_t arglen)
{
    hash_include_h inc_arg;
    assert(arg);
    assert(arglen);
    inc_arg = symbol_new(arg,arglen);
    return inc_arg;
}

int
hash_include_compare(hash_include_const_h lhs, void const * rhs, size_t arglen)
{
    int cmp;
    assert(lhs);
    assert(rhs);
    if (arglen) {
        cmp = STRNICMP(lhs->sym_name,rhs,arglen);
        if (!cmp && lhs->sym_name[arglen]) {
            cmp = 1;
        }
    } else {
        hash_include_const_h rarg = rhs;
        cmp = STRICMP(lhs->sym_name,rarg->sym_name);
    }
    return cmp;
}

char const *
hash_include_parse(char const *cp, char const **pdiagnostic, canonical_string_h *presolved_val)
{
    char const *arg_start = cp;
    char const *arg_end;
    assert(cp);
    assert(*presolved_val == NULL);
    cp = arg_end = chew_symbol(arg_start);
    if (arg_end == arg_start) { /* Argument is not a symbol */
        return hash_include_parse_final(cp,pdiagnostic);
    } else { /* Argument may be symbol or macro call */
        cp = chew_macro_call(arg_end);
        if (cp > arg_end) { /* Is a macro call */
            if (cp[-1] != ')') { /* Is incomplete macro call */
                *pdiagnostic = "unclosed macro-call";
            } else {
                *pdiagnostic = "macro-expansion is not supported";
            }
        } else { /* Argument is a symbol */
            /* Is it a configured symbol? */
            canonical_string_h symname = canonical_identifier(&arg_start);
            symbol_h  sym = configured_symbol_match(symname);
            if (sym) { /* Is a configured symbol */
                canonical_string_const_h final_val = NULL;
                if (eligible_line(CMD_SYMBOLS,HASH_INCLUDE)) {
                    configured_symbol_tally(sym);
                }
                final_val = symbol_final_value(sym);
                *presolved_val = canonical_string_copy(final_val);
                if (final_val) { /* Configured symbol is -Ded */
                    /* Evaluate the symbol's final value */
                    (void)hash_include_parse_final(canonical_string_text(final_val),pdiagnostic);
                } else { /* Configured symbol is -Ued */
                    *pdiagnostic = "argument resolves to 0";
                }
            } else {
                *pdiagnostic = "argument is an unknown symbol";
                if (eligible_line(CMD_SYMBOLS,HASH_INCLUDE)) {
                    unconfigured_symbol_tally(canonical_string_copy(symname));
                }
            }
            canonical_string_dispose(symname);
        }
    }
    return cp;
}


void
hash_include_define(hash_include_h hi, canonical_string_const_h def)
{
    assert(hi);
    if (hi->sym_def) {
        canonical_string_assign(hi->sym_def,def);
    } else {
        hi->sym_def = canonical_string_copy(def);
    }
}
/* EOF */





