#ifndef HASH_INCLUDE_H
#define HASH_INCLUDE_H
/***************************************************************************
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "opts.h"
#include "symbol.h"

/*!\ingroup hash_include_component hash_include_interface
	\file hash_include.h
	This file provides the Include-Argument component interface
*/

/*! \addtogroup hash_include_interface*/
/*@{*/


/*!	Abstract type of argument to a hash-include directive.

	Because the argument of an hash-include may be a defined preprocessor
	symbol we just make this type an alias for \e symbol_h
*/
typedef symbol_h hash_include_h;

/*!	Abstract type of const argument to hash-include directive */
typedef symbol_const_h hash_include_const_h;

/*! Construct a new hash-include argument object.
	\param	arg	The start of the hash-include argument
	\param	arglen	The length of \e arg.
*/
extern hash_include_h
hash_include_new(char const *arg, size_t arglen);

/*! Destroy an hash_include object.
	The destructor is the same as for \e symbol_h
*/
#define hash_include_dispose	symbol_dispose

/*! Construct a copy of an hash_include object
	The function is the same as for \e symbol_h
*/
#define hash_include_copy	symbol_copy

/*! Exchange two hash_include objects
	The function is the same as for \e symbol_h
*/
#define hash_include_swap	symbol_swap

/*! Assign one hash_include object to another.
	The function is the same as for \e symbol_h
*/
#define hash_include_assign	symbol_assign

/*! Get the definition of an hash_include.
	The function is the same as for \e symbol_h
*/
#define hash_include_definition	symbol_definition

/*! Undefine an hash_include.
	The function is the same as for \e symbol_h
*/
#define hash_include_undefine symbol_undefine

/*! Get the argument string from a hash-include argument.
	\param	inc_arg The hash-include argument to be queried.
	\return	Pointer to the 0-terminated argument string.
*/
#define hash_include_str(inc_arg) symbol_name(inc_arg)

/*!	Say whether a hash-include argument has been seen before
	The function is the same as for \e symbol_h.
*/
#define hash_include_seen symbol_seen

/*! Mark a hash-include argument as seen.
	The function is the same as for \e symbol_h.
*/
#define hash_include_memorise symbol_memorise

/*!	Forget that a hash-include argument has been seen.
	The function is the same as for \e symbol_h.
*/
#define hash_include_forget symbol_forget


/*! Assign a definition to a hash_include.
	\param	hi	The hash-include to which the definition shall be assigned.
	\param	def	A canonical string containing the definition to assign.
*/
extern void
hash_include_define(hash_include_h hi, canonical_string_const_h def);

/*! Compare an \e hash_include_h with another one or with a hash-include argument.
	\param	lhs	 First  \e hash_include_h to be compared.
	\param	rhs	 Second \e hash_include_h or hash-include argument to compare.
	\param	namelen	If 0 then \e rhs addresses a second \e hash_include_h to be
			compared by name with \e lhs. Otherwise \e rhs addresses a
			hash-include argument of length \e namelen to be compared with the
			name of \e lhs.
	\return A negative value if \e lhs is less than \e rhs; 0 if \e lhs
		is equal to \e rhs; a positive value if \e lhs is greater than \e rhs.
*/
extern int
hash_include_compare(hash_include_const_h lhs, void const * rhs,
                     size_t namelen);


/*!	Parse the argument of a hash-include directive.
	\param cp	The start of the hash-include argument.
	\param pdiagnostic	Pointer at which will be stored on return either
        NULL or the address of a diagnostic. If NULL is stored then
        the argument of the hash-include directive is valid; otherwise
        the stored diagnostic says why it is problematic.
    \param presolved_val On entry, a pointer to NULL. Onn return
        either NULL or a canonical string will be stored here. NULL is stored
        if the argument of the directive is not a symbol, or is a -U configured symbol
        or an unknown symbol. If a canonical string is stored then the
        argument is a -D-configured symbol and the stored canonical string
        holds the definition of that symbol, which may be an empty string.
    \return The address of the first unparsed character.

        It is the caller's responsibility to destroy any canonical
        strings stored at \c presolved_val.

*/
extern char const *
hash_include_parse( char const *cp,
                    char const **pdiagnostic,
                    canonical_string_h *presolved_val);

/*!	Say whether an hash-include argument has evaluated differently from
	its last evaluation.
	\param	inc_arg		The hash-include argument to evaluate.
	\return	True if the most recent evaluation of \e inc_arg has flagged
	it as evaluating differently from the evaluation before that, provided
	that \e hash_include_update_status() has not been called since the
	latest evaluation. The first evaluation of \e inc_arg will always flag
	it as having a new evaluation.
*/
extern bool
hash_include_changed(hash_include_const_h inc_arg);

/*!	Accept the most recent evaluation of an hash-include argument as the
	baseline evaluation.
	\param inc_arg	The hash-include argument to be updated.

	The function clears the flag evaluated by \e hash_include_changed(),
	so that the last evaluation becomes the basis of comparision
	for subsequent ones.
*/
extern void
hash_include_commit_evaluation(hash_include_h inc_arg);

/*!	Say whether an hash-include argument is valid.
	\param inc_arg	The hash-include argument to be queried.

	\e inc_arg is valid if it represents either a valid header name
	or a preprocessor symbol that finally evaluates to a valid
	header name.
*/
extern bool
hash_include_valid(hash_include_const_h inc_arg);

/*@}*/


#endif /* EOF*/

