/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "integer_constant.h"
#include "report.h"
#include "chew.h"
#include <stdio.h>
#include <limits.h>



/*!\ingroup integer_constant_component integer_constant_interface integer_constant_internals
 *\file integer_constant.c
 * This file implements the Integer Constant component
*/

/*! \addtogroup integer_constant_internals*/
/*@{*/

/* Helpers */

/*! Consume a possibly escaped ASCII character from a buffer returning its escaped value.
    \param  cpp  Pointer to the address from which a character is to be consumed. On return,
        the address of the next unconsumed character is stored here. A null character
        will not be consumed.
    \return The integer value of the possibly escaped character consumed, if any, else EOF.

    If \b cpp addresses the escape character '\\' on entry then that character is
    consumed and, if the next character is non-null, its escaped value is returned: if the next character
    is null then EOF is returned.

    If \b cpp does not address the escape character on entry then, if \b *cpp is non-null, its
    integer value is returned: if \b *cpp is null then EOF is returned.
*/
static int
char_escape_val(char const **cpp)
{
    int val = EOF;
    char const *cp = *cpp;
    if (*cp == '\\') {
        cp = chew_continuation(++cp);
        switch(*cp) {
        case '\0':
            break;
        case 'a':
            val = '\a';
            break;
        case 'b':
            val = '\b';
            break;
        case 'f':
            val = '\f';
            break;
        case 'n':
            val = '\n';
            break;
        case 'r':
            val = '\r';
            break;
        case 't':
            val = '\t';
            break;
        case 'v':
            val = '\v';
            break;
        default:
            val = *cp;
        }
        cp = chew_continuation(++cp);
    } else if (*cp) {
        val = *cp;
        cp = chew_continuation(++cp);
    }
    *cpp = cp;
    return val;
}

/*! Decode an UTF-8 encoded character from a buffer, ignoring
    embedded line-continautions.
    \param  cpp On entry, a pointer to the address of the first byte to
        to be decoded. On return, stores the address of the first byte
        not decoded.
    \return The integer value of the UTF-8 character parsed, or ULLONG_MAX
        if none is parsed.

*/
static unsigned long long
decode_utf8(char const **cpp)
{
    unsigned long long val = ULLONG_MAX;
    char const *cp = *cpp;
    size_t bytes = 0;
    unsigned int byte1 = *cp & 0xff;
    unsigned int byte2, byte3, byte4;
    if (byte1 == EOF) {
        bytes = 0;
    } else if (byte1 >> 7 == 0) {
        /* Top bit of lead byte is 0. This is a plain ascii code */
        bytes = 1;
    } else if (byte1 >> 5 == 0x6) {
        /* Top 3 bits of lead byte are 110. A 2-byte code. */
        bytes = 2;
    } else if (byte1 >> 4 == 0xe) {
        /* Top 4 bits of lead byte are 1110. A 3-byte code. */
        bytes = 3;
    } else if (bytes >> 3 == 0x1e) {
        /* Top 5 bits of lead byte are 11110. A 4-byte code. */
        bytes = 4;
    }
    switch(bytes) {
    case 1:
        val = byte1;
        *cpp = chew_continuation(++cp);
        break;
    case 2:
        byte1 &= 0x1F; /* Clear top three bits of lead byte. */
        byte2 = 0xff & *(cp = chew_continuation(++cp));
        if (byte2 == EOF || byte2 >> 6 != 2) { /* Top 2 bits of continuation byte must be 10 */
            break;
        }
        byte2 &= 0x3F;
        /* Value is concatenation of non-control bits */
        val = ((byte1 << 6) | byte2);
        *cpp = chew_continuation(++cp);
        break;
    case 3:
        byte1 &= 0x1F;
        byte2 = 0xff & *(cp = chew_continuation(++cp));
        if (byte2 == EOF || byte2 >> 6 != 2) {
            break;
        }
        byte2 &= 0x3F;
        byte3 = 0xff & *(cp = chew_continuation(++cp));
        if (byte3 == EOF || byte3 >> 6 != 2) {
            break;
        }
        byte3 &= 0x3F;
        val = (byte1 << 12) | (byte2 << 6) | byte3;
        *cpp = chew_continuation(++cp);
        break;
    case 4:
        byte1 &= 0x1F;
        byte2 = 0xff & *(cp = chew_continuation(++cp));
        if (byte2 == EOF || byte2 >> 6 != 2) {
            break;
        }
        byte2 &= 0x3F;
        byte3 = 0xff & *(cp = chew_continuation(++cp));
        if (byte3 == EOF || byte3 >> 6 != 2) {
            break;
        }
        byte3 &= 0x3F;
        byte4 = 0xff & *(cp = chew_continuation(++cp));
        if (byte4 == EOF || byte4 >> 6 != 2) {
            break;
        }
        byte4 &= 0x3F;
        val = (byte1 << 18) | (byte2 << 12) | (byte3 << 6) | byte4;
        *cpp = chew_continuation(++cp);
        break;
    default:
        break;
    }
    return val;
}

/*! Get the integer value of a numeral.
    \param  pnum On entry, pointer to the address of the numeral to evaluate.
                On return, the address of the first unconsumed character.
    \param  base The arithmetic base in which the numeral is to be evaluated.
    \param  overflow On return, \b true is stored here is the integer value of the
        numeral overflows unsigned long long, else \b false is stored.
    \return The integer value of the numeral as unsigned long long,
        if it is representable as such. The returned value is not
        valid if overflow occurred/

    The function consumes characters after \b num for as long as
    they can form part of a numeral with the indicated base.
*/
static unsigned long long
eval_numeral_helper(char const **pnum, unsigned int base, bool *overflow)
{
    char const *num = *pnum;
    int dval = 0;
    unsigned long long val = 0;
    unsigned long long tmp;
    *overflow = false;
    for (	;; num = chew_continuation(++num)) {
        switch(*num) {
        case '0':
            dval = 0;
            break;
        case '1':
            dval = 1;
            break;
        case '2':
            dval = 2;
            break;
        case '3':
            dval = 3;
            break;
        case '4':
            dval = 4;
            break;
        case '5':
            dval = 5;
            break;
        case '6':
            dval = 6;
            break;
        case '7':
            dval = 7;
            break;
        case '8':
            dval = 8;
            break;
        case '9':
            dval = 9;
            break;
        case 'a':
        case 'A':
            dval = 10;
            break;
        case 'b':
        case 'B':
            dval = 11;
            break;
        case 'c':
        case 'C':
            dval = 12;
            break;
        case 'd':
        case 'D':
            dval = 13;
            break;
        case 'e':
        case 'E':
            dval = 14;
            break;
        case 'f':
        case 'F':
            dval = 15;
            break;
        default:
            dval = 16;
        }
        if (dval < base) {
            tmp = val;
            val *= base;
            if (val / base != tmp) {
                *overflow = true;
            } else if (dval) {
                tmp = val;
                val += dval;
                if (val <= tmp) {
                    *overflow = true;
                }
            }
        } else {
            break;
        }
    }
    *pnum = num;
    return val;
}


/*! Consume a numerically encoded character constant from a buffer returning its
    integer value.
    \param  cpp Pointer to the address from which the constant is to be consumed. On
        return, the address of the next unconsumed character is stored here.
    \param base The presumed base of the constant, either 8 (octal) or 16 (hex).
    \return The integer value of encoded constant, if any is found,
    else ULONG_MAX.

    It is assumed that the possible constant at \b *cpp is enclosed in
    single quotes and that \b *cpp on entry addresses the
    opening quote. If parsing of an encoded constant does not conume a closing quote
    then nothing is consumed from the buffer
*/
static unsigned long long
eval_numerical_character_constant(char const **cpp, int base)
{
    unsigned long long val = ULLONG_MAX;
    char const *cp = *cpp;
    char prefix = 0;
    assert(base == 8 || base == 16);
    if (base == 8) {
        prefix = '0';
    } else if (base == 16) {
        prefix = 'x';
    } else {
        assert(false);
    }
    if (*cp == '\'') {
        cp = chew_continuation(++cp);
        if (*cp == '\\' && *(cp = chew_continuation(++cp)) == prefix) {
            bool overflow;
            cp = chew_continuation(++cp);
            val = eval_numeral_helper(&cp,base,&overflow);
            if (overflow || val > INT_MAX || *cp != '\'') {
                val = ULLONG_MAX;
            } else {
                *cpp = chew_continuation(++cp);
            }
        }
    }
    return val;
}

/*! Consume a an octal character constant from a buffer returning its
    integer value.
    \param  cpp Pointer to the address from which the constant is to be consumed. On
        return, the address of the next unconsumed character is stored here.
    \return The integer value of the octal constant, if an octal constant is found,
    else ULONG_MAX.

    It is assumed that the possible octal constant at \b *cpp is enclosed in
    single quotes and that \b *cpp on entry addresses the
    opening quote. If parsing of an octal constant does not conume a closing quote
    then nothing is consumed from the buffer
*/
static unsigned long long
eval_octal_character_constant(char const **cpp)
{
    return eval_numerical_character_constant(cpp,8);
}

/*! Consume a hex character constant from a buffer returning its
    integer value.
    \param  cpp Pointer to the address from which the constant is to be consumed. On
        return, the address of the next unconsumed character is stored here.
    \return The integer value of the hex constant, if a hex constant is found,
    else ULONG_MAX.

    It is assumed that the possible hex constant at \b *cpp is enclosed in
    single quotes and that \b *cpp on entry addresses the
    opening quote. If parsing of a hex  constant does not conume a closing quote
    then nothing is consumed from the buffer
*/
static unsigned long long
eval_hex_character_constant(char const **cpp)
{
    return eval_numerical_character_constant(cpp,16);
}

/*@}*/

/* API*/

int_spec_t
eval_numeral(char const *num, char const **numend)
{
    int_spec_t result = {INT_UNION_INITOR,INT_INSOLUBLE};
    unsigned int base = 10;
    unsigned long long val, max_val = ULLONG_MAX;
    bool overflow = false;
    int_type type = INT_LLONG;
    /*  Initially assume numeral may represent a long long
        until a type suffix confirms or disconfirms.
        Finally, in the absence of any type suffix, we will say the numeral
        is an int (signed or unsigned) if it is not too big */
    size_t num_len = 0;
    char const *start = num;
    char const *int_type = NULL;
    heap_str gripe = NULL;
    bool has_suffix = false;
    if (*num == '0') {
        num = chew_continuation(++num);
        if (*num == 'x' || *num == 'X') {
            num = chew_continuation(++num);
            base = 16;
        } else {
            base = 8;
        }
    }
    val = eval_numeral_helper(&num,base,&overflow);
    if (overflow) {
        type = INT_ULLONG;
        int_type = "unsigned long long";
        max_val = ULLONG_MAX;
    }
    num_len = num - start;
    if (num_len == 1 && (*num == 'x' || *num == 'X')) {
        num = start;
    } else if (num_len > 0) {
        if (*num == 'u' || *num == 'U') {
            has_suffix = true;
            type = INT_UINT;
            num = chew_continuation(++num);
        }
        if (*num == 'l' || *num == 'L') {
            has_suffix = true;
            num = chew_continuation(++num);
            type = IS_SIGNED(type) ? INT_LONG : INT_ULONG;
        }
        if (*num == 'l' || *num == 'L') {
            has_suffix = true;
            num = chew_continuation(++num);
            type = IS_SIGNED(type) ? INT_LLONG : INT_ULLONG;
        }
        if (IS_SIGNED(type) && (*num == 'u' || *num == 'U')) {
            if (!has_suffix) {
                has_suffix = true;
                type = INT_UINT;
            } else {
                type = MAKE_UNSIGNED(type);
            }
            num = chew_continuation(++num);
        }

    }
    *numend = num;
    if (num_len == 0) {
        return result;
    }
    if (IS_SIGNED(type) && val > LLONG_MAX && !has_suffix) {
        /* If we have no suffix explicitly specifying long or long long
            then we allow the value to be an unsigned long long and emit a warning.
            We store the warning at this point and will write it later if it
            is not superceded by an overflow warning */
        report(GRIPE_FORCED_UNSIGNED,&gripe,
               "Integer constant \"%.*s\" is so large it is treated as unsigned",
               num - start,start);

        type = INT_ULLONG;
    }
    if (type == INT_ULONG && val > ULONG_MAX) {
        overflow = true;
        int_type = "unsigned long";
        max_val = ULONG_MAX;
    } else if (type == INT_LONG && val > LONG_MAX) {
        overflow = true;
        int_type = "long";
        max_val = LONG_MAX;
    } else if (type == INT_LLONG && val > LLONG_MAX) {
        overflow = true;
        int_type = "long long";
        max_val = LLONG_MAX;
    }
    if (overflow) {
        release((void **)&gripe); /* On overflow discard any stored warning */
        report(GRIPE_INT_OVERFLOW,NULL,
               "Integer constant \"%.*s\" is too large for type %s "
               "(max %llu): expression will not be resolved",
               num - start,start,int_type,max_val);
        result.type = INT_TOO_BIG;
    } else {
        if (gripe) { /* Didn't overflow so emit any stored warning */
            report(GRIPE_FORCED_UNSIGNED,&gripe,NULL);
        }
        if (!has_suffix) {
            /* No type-suffix. Reduce type to (unsigned) int if value will fit */
            if (IS_SIGNED(type)) {
                if (val <= (unsigned long long)INT_MAX) {
                    type = INT_INT;
                }
            } else if (val <= (unsigned long long)UINT_MAX) {
                type = INT_UINT;
            }
        }
        result.val.ull = val;
        result.type = type;
    }
    return result;
}

/*! Consume an character constant from a buffer, returning
    its integer value as an \em int_spect_t.
    \param  cpp  Pointer to the address of the putative character constant.
        On return, the address of the next unconsumed character is stored
        here.
    \return An \e int_spec_t that represents the integer value of
        the parsed character constant, or INT_SPEC_INITOR
        if no constant is parsed.
*/
int_spec_t
eval_character_constant(char const **cpp)
{
    int_spec_t result = INT_SPEC_INITOR;
    char const *cp = *cpp;
    char const *int_type = "int";
    unsigned long long val = ULLONG_MAX;
    bool wide = false;
    int max_len = sizeof(int); /* Assume multi-byte constant. Can have sizeof(int) bytes */
    if (*cp == 'L') { /* No, this is a wide-character constant */
        cp = chew_continuation(++cp);
        max_len = sizeof(wchar_t); /* Can have sizeof(wchar_t) bytes */
        int_type = "wchar_t";
        wide = true;
    }
    if (*cp == '\'') {
        if (val == ULLONG_MAX) {
            val = eval_hex_character_constant(&cp);
        }
        if (val == ULLONG_MAX) {
            val = eval_octal_character_constant(&cp);
        }
        if (val == ULLONG_MAX) {
            /* Not a hex or octal constant. Try multi-byte or wide constant */
            int chval = 0;
            int nbytes = 0;
            val = 0;
            cp = chew_continuation(++cp); /* Consume opening quote */
            /* If not a wide character constant then we'll evaluate as a concatenation of
                8-bit ints. But likewise even if it is a wide character constant that commences with
                an escape, because an escaped character must be ascii. */
            if (!wide || (cp[0] == '\\' && cp[1] != '\n')) {
                for (; nbytes < max_len && *cp && *cp != '\'' && (chval = char_escape_val(&cp)) != EOF;
                        val = (val << 8 | chval), ++nbytes) {}
            } else {
                val = decode_utf8(&cp);
            }
            if (val == ULLONG_MAX) {}
            if (*cp != '\'') { /* Not a valid constant. No closing quote within max_len. */
                val = ULLONG_MAX;
                /* Look for that closing quote... */
                for (; *(cp = chew_continuation(++cp)) != '\0' && *cp != '\n' && *cp != '\'';
                        cp = chew_continuation(++cp)) {};
                if (!*cp || *cp == '\n') { /* Unclosed quotation */
                    report(GRIPE_UNCLOSED_QUOTE,NULL,
                           "Unclosed quotation after \"%.*s\"",cp - *cpp,*cpp);
                } else { /* Constant is too long for type */
                    report(GRIPE_CHAR_CONSTANT_TOO_LONG,NULL,
                           "Character constant %.*s is too long for type %s (max %d bytes). Will not be resolved",
                           (cp - *cpp) + 1,*cpp,int_type,max_len);
                }
            } else if (!wide && nbytes > 1) {
                report(GRIPE_MULITBYTE_CHAR_CONSTANT,NULL,
                       "Multi-byte character constant %.*s",(cp - *cpp) + 1,*cpp);
            }
            cp = chew_continuation(++cp);
        }
    }
    if (val != ULLONG_MAX) {
        result.val.i = val;
        result.type = INT_INT;
    } else {
        result.type = INT_INSOLUBLE;
    }
    *cpp = cp;
    return result;
}


/* EOF */
