#ifndef INTEGER_OPS_H
#define INTEGER_OPS_H
/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/


/*!\ingroup integer_ops_component integer_ops_interface
 *\file lexicon.h
 * This file provides the Integer Operations interface.
*/

/*!	\addtogroup integer_ops_interface */
/*@{*/

/*! Encoding of integer types.
    The upper 3rd byte encodes the integer conversion rank of the integer type.
        See standard ISO/IEC 9899:1999 6.3.1.1
    The upper 2nd byte contains the size of the type.
    The lowest byte encodes the signedness of the type, 0 = unsigned, 1 = signed.

*/
typedef enum {
    INT_UNDEF = 0,                                      /*!< No defined type */
    INT_CHAR = (1 << 16) | (sizeof(char) << 8) | 1,     /*!< char */
    INT_UCHAR = (1 << 16) | sizeof(char) << 8,          /*!< unsigned char */
    INT_SHORT = (2 << 16) | (sizeof(short) << 8) | 1,    /*!< short int */
    INT_USHORT = (2 << 16) | sizeof(short) << 8,       /*!< unsigned short int */
    INT_INT = (3 << 16) | (sizeof(int) << 8) | 1,        /*!< int */
    INT_UINT = (3 << 16) | sizeof(int) << 8,           /*!< unsigned int */
    INT_LONG = (4 << 16) | (sizeof(long) << 8) | 1,      /*!< long int */
    INT_ULONG = (4 << 16) | sizeof(long) << 8,         /*!< unsigned long int */
    INT_LLONG = (5 << 16) | (sizeof(long long) << 8) | 1,     /*!< long long int */
    INT_ULLONG = (5 << 16) | sizeof(long long) << 8,        /*!< unsigned Long long int */
    INT_TOO_BIG = 6 << 16,       /*!< Integer value too large for its type */
    INT_INSOLUBLE = INT_TOO_BIG /*!< Thing was not an int at all */
} int_type;

/*! Get the integer conversion rank from an \e int_type */
#define RANK(int_type) ((int_type) >> 16)

/*! Get the size of an integer type from an \e int_type */
#define SIZEOF(int_type) (((int_type) >> 8) & ~0xff00)

/*! Say whether an \e int_type represents a signed type */
#define IS_SIGNED(int_type) ((int_type) & 1)

/*! Get the signed variant of an \e int_type */
#define MAKE_SIGNED(int_type) ((int_type) | 1)

/*! Get the unsigned variant of an \e int_type */
#define MAKE_UNSIGNED(int_type) ((int_type) & ~1)

/*! Structure representing the an integer value */
typedef struct int_spec {
    union {
        long long ll;
        unsigned long long ull;
        long l;
        unsigned long ul;
        int i;
        unsigned ui;
    } val; /*!< The value of the integer */
    int_type type; /*!< Type of the integer */
} int_spec_t;

/*! Initialiser for union of all integer types */
#define INT_UNION_INITOR {0LL}

/*! Initialiser for type int_spec_t */
#define INT_SPEC_INITOR {INT_UNION_INITOR,INT_UNDEF}

/*! Function type of binary operations on \em int_spec_t objects */
typedef int_spec_t (integer_bin_op_t)(int_spec_t const *, int_spec_t const *);

/*! less-than op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the comparison as an \em int_spec_t
*/
extern int_spec_t
lt(int_spec_t const * l, int_spec_t const *r);

/*! greater-than op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the comparison as an \em int_spec_t
*/
extern int_spec_t
gt(int_spec_t const * l, int_spec_t const *r);

/*! less-than-or-equal op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the comparison as an \em int_spec_t
*/
extern int_spec_t
le(int_spec_t const * l, int_spec_t const *r);

/*! greater-than-or-equal op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the comparison as an \em int_spec_t
*/
extern int_spec_t
ge(int_spec_t const * l, int_spec_t const *r);

/*! equality op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the comparison as an \em int_spec_t
*/
extern int_spec_t
eq(int_spec_t const * l, int_spec_t const *r);

/*! inequality op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the comparison as an \em int_spec_t
*/
extern int_spec_t
ne(int_spec_t const * l, int_spec_t const *r);

/*! bit_and op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the operation as an \em int_spec_t
*/
extern int_spec_t
bit_and(int_spec_t const * l, int_spec_t const *r);

/*! bit_inclusive_or op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the operation as an \em int_spec_t
*/
extern int_spec_t
bit_or(int_spec_t const * l, int_spec_t const *r);

/*! bit_exclusive_or op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the operation as an \em int_spec_t
*/
extern int_spec_t
bit_xor(int_spec_t const * l, int_spec_t const *r);

/*! complement op on an \em int_spec_t object.
    \param int_spec   The \em int_spec to be complemented.
    \return The result of the operation as an \em int_spec_t
*/
extern int_spec_t
complement(int_spec_t const * int_spec);

/*! unary_minus op on an \em int_spec_t object.
    \param int_spec   The \em int_spec to be negated.
    \return The result of the operation as an \em int_spec_t
*/
extern int_spec_t
minus(int_spec_t const * int_spec);

/*! left_shift op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the shift as an \em int_spec_t
*/
extern int_spec_t
lshift(int_spec_t const * l, int_spec_t const *r);

/*! right_shift op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the shift as an \em int_spec_t
*/
extern int_spec_t
rshift(int_spec_t const * l, int_spec_t const *r);

/*! addition op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the addition as an \em int_spec_t
*/
extern int_spec_t
add(int_spec_t const * l, int_spec_t const *r);

/*! subtraction op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the subtraction as an \em int_spec_t
*/
extern int_spec_t
subtract(int_spec_t const * l, int_spec_t const *r);

/*! multiply op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the multiplication as an \em int_spec_t
*/
extern int_spec_t
multiply(int_spec_t const * l, int_spec_t const *r);

/*! divide op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the division as an \em int_spec_t
*/
extern int_spec_t
divide(int_spec_t const * l, int_spec_t const *r);

/*! modulus op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the mod op as an \em int_spec_t
*/
extern int_spec_t
modulus(int_spec_t const * l, int_spec_t const *r);

/*! boolean_and op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the conjunction as an \em int_spec_t
*/
extern int_spec_t
boolean_and(int_spec_t const * l, int_spec_t const *r);


/*! boolean_or op on two \em int_spec_t objects.
    \param l    The left hand operand
    \param r    The right hand operand.
    \return The result of the conjunction as an \em int_spec_t
*/
extern int_spec_t
boolean_or(int_spec_t const * l, int_spec_t const *r);


/*! boolean_negation op on an \em int_spec_t object.
    \param int_spec   The \em int_spec to be negated.
    \return The result of the negation as an \em int_spec_t
*/
extern int_spec_t
negate(int_spec_t const * int_spec);

/*@}*/

#endif /* EOF*/


