/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "lexicon.h"
#include "chew.h"
#include "symbol.h"
#include <string.h>
#include <stddef.h>

/*!\ingroup lexicon_component lexicon_interface lexicon_internals
 *\file lexicon.c
 * This file implements the Lexicon component
*/

/*! \addtogroup lexicon_internals*/
/*@{*/

/* Helpers */

/*! Compare text in the line buffer with a reference string, ignoring line-continuations
    in the line buffer.
    \param ptxt	    On entry, a pointer to an address in line buffer at which to compare text to the reference string.
                    On return, stores the address of the first byte in the line buffer not consumed by the
                    in the matched text, if the text is matched within \c nbytes bytes; otherwise unchanged.
    \param match	The reference string to match.
    \param nbytes   The maximum number of bytes that may be compared. If 0 then
                    defaults to \c strlen(match).
    \return		    0 if the text compares lexicographically equal to the reference string within \c nbytes bytes.
                    1 the text compares greater, -1 if the text compares less.
*/
static int
text_match(char const **ptxt, char const *match, size_t nbytes)
{
    int cmp;
    char const *txt = *ptxt;
    if (!nbytes) {
        nbytes = strlen(match);
    }
    for ( ; --nbytes && *txt && *match && *txt == *match; txt = chew_continuation(++txt),++match) {}
    if (*txt < *match) {
        cmp = -1;
    } else if (*txt > *match) {
        cmp = 1;
    } else {
        cmp = 0;
        *ptxt = chew_continuation(++txt);
    }
    return cmp;
}

/*@}*/

/* API*/

/*! Say whether text in the line-buffer matches a prepprocessor keyword
    \param  ptxt    On entry, a pointer to an address in the line buffer at which to
            match the keyword. On return, stores the address of the first byte in the line buffer not consumed by the
            in the matched text, if the text is an unambiguous occurrence of the keyword; otherwise unchanged.
    \param  keyword The keyword to be matched.
    \return True if the text at *ptxt on entry is an unambiguous occurrence of \c keyword; else false.

*/
bool
match_keyword(char const **ptxt, char const *keyword)
{
    char const *saved_txt = *ptxt;
    if (0 == text_match(ptxt,keyword,0) && !is_symbol_start_char(**ptxt)) {
        return true;
    } else {
        *ptxt = saved_txt;
        return false;
    }
}

bool
match_op(char const ** ptxt, char const *op)
{
    char const *txt = *ptxt;
    bool verdict = true;
    assert(op[1] == '\0' || op[2] =='\0');
    if (txt[0] != op[0]) {
        verdict = false;
    } else {
        if (op[1]) {
            txt = chew_continuation(++txt);
            if (txt[0] != op[1]) {
                verdict = false;
            }
        } else if (txt[1] == op[0]) {
            verdict = false;
        }
    }
    if (verdict) {
        *ptxt = txt + 1;
    }
    return verdict;
}

/* EOF*/

