/***************************************************************************
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 ***************************************************************************/

#include "ptr_set.h"
#include "memory.h"
#include "swiss_army.h"
#include <string.h>
#include <assert.h>


/*! \ingroup ptr_set_component ptr_set_interface ptr_set_internals
	\file ptr_set.c
    This file implements the Pointer-Set component
*/

/*!	\addtogroup ptr_set_internals*/
/*@{*/

/*! Structure of a pointer set */
struct ptr_set {
    /*!	Underlying pointer vector */
    ptr_vector_h pv;
    /*! Comparator function */
    comparator_t compare;
    /*! Cached position at which the last key unsuccessfuly searched for
    	should be inserted.
    */
    locator_t cached_pos;
};

/* Helpers */

/*!	Default comparator function for ordering pointer sets.
	\param	obj	Address to an object to be compared with a key.
	\param	key	Address of the key to compare with \e obj.
	\param	ignore Unused parameter required by \e comparator_t prototype.
	\return The result of \e strcmp(obj,key).
*/
static int
default_comparator(void const *obj, void const *key, locator_t ignore)
{
    (void)ignore;
    return strcmp(obj,key);

}

/*!	Initialise a pointer set.
	\param	ps	The pointer set to initialise. Is assumed to control
				only unitialised storage.
	\param	compare	The comparator to be used for ordering this pointer set.
	\param	dtor The destructor to be applied to members of this pointer set.
	\param	cloner The function for cloning elements of this pointer set.
*/
static void
ptr_set_init(	ptr_set_h ps,
                comparator_t compare,
                dtor_t dtor,
                cloner_t cloner)
{
    ps->pv = ptr_vector_new(dtor,cloner);
    ps->compare = compare ? compare : default_comparator;
    ps->cached_pos = -1;
}

/*!	Initialise a pointer set from another
	\param	dest The pointer set to initialise. Is assumed to control
				only unitialised storage.
	\param	src	The pointer set to be copied.

	\e dest is initialised as a copy of \e src
*/
static void
ptr_set_copy_init(ptr_set_h dest, ptr_set_const_h src)
{
    dest->pv = ptr_vector_copy(src->pv);
    dest->compare = src->compare;
    dest->cached_pos = src->cached_pos;
}

/*@}*/

/* API **************************************************************/

ptr_set_h
ptr_set_new(comparator_t compare, dtor_t dtor, cloner_t cloner)
{
    ptr_set_h ps = allocate(sizeof(struct ptr_set));
    ptr_set_init(ps,compare,dtor,cloner);
    return ps;
}

void
ptr_set_dispose(ptr_set_h ps)
{
    if (ps) {
        ptr_vector_dispose(ps->pv);
        free(ps);
    }
}

ptr_set_h
ptr_set_copy(ptr_set_const_h src)
{
    ptr_set_h ps;
    assert(src);
    ps = allocate(sizeof(struct ptr_set));
    ptr_set_copy_init(ps,src);
    return ps;
}

void
ptr_set_swap(ptr_set_h lhs, ptr_set_h rhs)
{
    assert(lhs);
    assert(rhs);
    ptr_vector_swap(lhs->pv,rhs->pv);
    PODSWAP(comparator_t,&lhs->compare,&rhs->compare);
    PODSWAP(locator_t,&lhs->cached_pos,&rhs->cached_pos);
}

void
ptr_set_assign(ptr_set_h dest, ptr_set_const_h src)
{
    if (dest != src) {
        ptr_set_h tmp = ptr_set_copy(src);
        ptr_set_swap(dest,tmp);
        ptr_set_dispose(tmp);
    }
}

bool
ptr_set_equal(	ptr_set_const_h lhs,
                ptr_set_const_h rhs)
{
    bool eq = lhs == rhs;
    if (!eq) {
        eq = lhs->compare == rhs->compare;
        if (eq) {
            eq = ptr_vector_equal(lhs->pv,rhs->pv,lhs->compare);
        }
    }
    return eq;
}


comparator_t
ptr_set_comparator(ptr_set_const_h ps)
{
    assert(ps);
    return ps->compare;
}

void *
ptr_set_search(ptr_set_h ps, char const *key, locator_t keylen)
{
    locator_t loc;
    assert(ps);
    loc = ptr_vector_search(ps->pv,key,keylen,ps->compare,true);
    if (NOWHERE(loc)) {
        ps->cached_pos = ~loc;
        return NULL;
    }
    ps->cached_pos = -1;
    return ptr_vector_at(ps->pv,(size_t)loc);
}

void const *
ptr_set_search_const(ptr_set_const_h ps, char const *key, locator_t keylen)
{
    return ptr_set_search((ptr_set_h)ps,key,keylen);
}


locator_t
ptr_set_insert(ptr_set_h ps, void const *obj)
{
    assert(ps);
    assert(obj);
    size_t nelems = ptr_vector_count(ps->pv);
    locator_t loc = ps->cached_pos;
    ps->cached_pos = -1;
    if (!nelems) {
        ptr_vector_append(ps->pv,(void *)obj);
    } else {
        if (SOMEWHERE(loc)) {
            int prev_comp = -1, next_comp = 1;
            void const *other;
            if (loc > 0) {
                other = ptr_vector_at(ps->pv,loc - 1);
                prev_comp = ps->compare(other,obj,0);
                if (!prev_comp) {
                    return ~(locator_t)other;
                }
            }
            if (loc + 1 < nelems) {
                other = ptr_vector_at(ps->pv,loc + 1);
                next_comp = ps->compare(other,obj,0);
                if (!next_comp) {
                    return ~(locator_t)other;
                }
            }
            if (prev_comp > 0 || next_comp < 0) {
                loc = -1;
            }
        }
        if (NOWHERE(loc)) {
            loc = ptr_vector_search(ps->pv,obj,0,ps->compare,true);
            if (SOMEWHERE(loc)) {
                return ~(locator_t)ptr_vector_at(ps->pv,loc);
            }
            loc = ~loc;
        }
        ptr_vector_insert(ps->pv,loc,obj);
    }
    return (locator_t)obj;
}

