#ifndef PTR_SET_H
#define PTR_SET_H
/***************************************************************************
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 ***************************************************************************/

#include "opts.h"
#include "ptr_vector.h"

/*!\ingroup ptr_set_component ptr_set_interface
 *\file ptr_set.h
 * This file provides the Pointer Set component interface
*/

/*! Abstract type of  pointer set */
typedef struct ptr_set * ptr_set_h;

/*! Abstract type of const pointer set */
typedef struct ptr_set const * ptr_set_const_h;


/*! Create new pointer set.
	\param	compare	The key comparison function for this pointer set, or NULL.
					If NULL then elements are compared by \e strcmp()
	\param	dtor	The destructor to be applied to set
					elements when the set is destroyed, or NULL. If
					NULL then \e free() is assumed.
	\param	cloner	The cloning function for cloning elements of this
					set, or NULL. If NULL then \e clone(ptr,0) is
					assumed.
	\return A new pointer set.
 */
extern ptr_set_h
ptr_set_new(comparator_t compare, dtor_t dtor, cloner_t cloner);

/*! Dispose of a pointer set.
	\param ps The pointer set to be destroyed.

	The set's \e dtor member is called on each element
    and then remaining memory controlled by the set is
    released.

	The function is a no-op if \e ps is NULL.
*/
extern void
ptr_set_dispose(ptr_set_h ps);

/*! Construct a copy of a pointer set.
	\param	src	The pointer set to copy
	\return A copy of \e src.
*/
extern ptr_set_h
ptr_set_copy(ptr_set_const_h src);

/*! Exchange two pointer sets */
extern void
ptr_set_swap(ptr_set_h lhs, ptr_set_h rhs);

/*! Assign one pointer set to another.
	\param	dest	The pointer set to be assigned
	\param	src		The pointer set to be assigned to \e dest.
*/
extern void
ptr_set_assign(ptr_set_h dest, ptr_set_const_h src);

/*! Say whether two pointer sets are equal.
	\param	lhs	The first pointer set to compare.
	\param	rhs	The second pointer set to compare.
	\return True if both sets are NULL or if they have the same
		comparator member and are equal in the same sense as
		pointer vectors with respect to that comparator.
*/
extern bool
ptr_set_equal(ptr_set_const_h lhs, ptr_set_const_h rhs);

/*! Get the comparison function a pointer set */
extern comparator_t
ptr_set_comparator(ptr_set_const_h ps);

/*! Get the destructor function a pointer set */
#define ptr_set_dtor(ps) ptr_vector_dtor((ptr_vector_const_h)ps)

/*! Get the cloning function from pointer set */
#define ptr_set_cloner(ps) ptr_vector_cloner((ptr_vector_const_h)ps)

/*! Get the number of elements in a pointer set */
#define ptr_set_count(ps) ptr_vector_count((ptr_vector_const_h)ps)

/*! Search for a key value in a pointer set.
	\param ps The pointer set to be searched.
	\param key	Start of the key value to be searched for.
	\param keylen	If postive this is the length of the key value
		 to be searched for. If 0 or negative then \e key is in fact
		the address of an object that is to be searched for by key.
	\return  If \e key is matched then the address of the
		matching element is returned; otherwise NULL.

	If \e key is not found then \e ps caches the position at which this key
	would sort in \e ps. If the next operation on \e ps
	is to insert an obj that matches \e key then insertion can be highly
	optimised by localising the second search of the set to the neighbours
 	of the cached position. If this optimisation fails because the object
	to be inserted next does not match the cached key then that object
	is inserted the long way.

	The pointer returned by this function is invalidated by any
	subsequent non-const operation on \e ps.

	If the return value is used to modify the key of the object it addresses
	then \e ps is corrupted.
*/
extern void *
ptr_set_search(ptr_set_h ps, char const *key, locator_t keylen);

/*! Const variant of \e ptr_set_search()
*/
extern void const *
ptr_set_search_const(ptr_set_const_h ps, char const *key, locator_t keylen);


/*! Insert an object into a pointer set.

	\param ps	The pointer set in which the object is to be inserted.
	\param	obj	Address of the object to be inserted,

	\return If \e obj is not already in the set then it is inserted and the
	index at which it is found is returned. Otherwise a negative value \e L
	is returned such that ~L is the index at which \e obj was found.

	If the return value is used to modify the key of the object it locates
	then \e ps is corrupted.
 */
extern locator_t
ptr_set_insert(ptr_set_h ps, void const *obj);


#endif /* EOF */
