#ifndef STATE_UTILS_H
#define STATE_UTILS_H
/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "bool.h"
#include "memory.h"
#include <stdlib.h>
#include <assert.h>
#include <string.h>

/*! \ingroup state_utils_interface state_utils_infrastructure
    \file state_utils.h
    The file provides the State Utils interface
*/

/*! \ingroup state_utils_infrastructure
 *
    These items that are imported by state_utils.h to implement
    the user macros but are inessential to client code.
*/
/*@{*/

#ifndef STATE_ALLOCATOR
/*! Define malloc-like function for allocating dynamic state storage*/
#define STATE_ALLOCATOR (malloc)
#endif

/*! Generated name for component global state instance*/
#define STATE(component)	component##_state
/*! Generated name for component public state instance*/
#define PUBLIC_STATE(component) component##_public_state
/*! Generated name for pointer to component global state instance */
#define HANDLE(component) component##_h
/*! Generated name for pointer to component public state instance*/
#define PUBLIC_HANDLE(component)	component##_public_h
/*! Mandatory name user defined component initialiser*/
#define USER_INIT(component)	component##_init
/*! Mandatory name user defined component finaliser*/
#define USER_FINIS(component)	component##_finis
/*! Mandatory name for pointer to user defined component initialiser*/
#define USER_INIT_HANDLE(component)	component##_init_h
/*! Mandatory name for pointer to user defined component finaliser*/
#define USER_FINIS_HANDLE(component)	component##_finis_h
/*! Generated name for component initialising function*/
#define INITOR(component)	component##_initor
/*!	Generated name for component finalising function*/
#define FINITOR(component)	component##_finitor
/*! Generated name for component static initialiser*/
#define STATIC_INITIALISER(component)	component##_static_initialiser
/*! Generated name for pointer to component static initialiser*/
#define STATIC_INITIALISER_HANDLE(component)	component##_static_initialiser_h
/*! enum tag name for component storage type, static or dynamic*/
#define STORAGE_TYPE_TAG(component)	component##_storage_type
/*! enum tag name for component's initialisation type*/
#define INITIALISATION_TYPE_TAG(component)		component##_initialisation_type
/*! Declare user-defined component initialising function*/
#define DECL_USER_INIT(component)\
	void USER_INIT(component)(STATE_T(component) *)
/*! Declare user-defined component finalising function*/
#define DECL_USER_FINIS(component)\
	void USER_FINIS(component)(STATE_T(component) *)
/*! Declare pointer to user-defined component initialising function*/
#define DECL_USER_INIT_HANDLE(component)\
	void (*USER_INIT_HANDLE(component))(STATE_T(component) *)
/*! Declare pointer to user-defined finalising function*/
#define DECL_USER_FINIS_HANDLE(component)\
	 void (*USER_FINIS_HANDLE(component))(STATE_T(component) *)
/*! Declare generated component initialising function*/
#define DECL_INITOR(component)	void INITOR(component)(void)
/*! Declare generated component finalising function*/
#define DECL_FINITOR(component)	void FINITOR(component)(void)
/*! Declare a static initialiser for the component*/
#define DECL_STATIC_INITIALISER(component)\
	static const STATE_T(component) STATIC_INITIALISER(component)
/*! Declare pointer to component const initialiser. Can be null if none*/
#define DECL_STATIC_INITIALISER_HANDLE(component)\
	static const STATE_T(component) * const STATIC_INITIALISER_HANDLE(component)
/*! Select static storage for component state instance*/
#define SELECT_STATIC(component)\
	enum component##_storage_type { STORAGE_TYPE_TAG(component) = STATE_STATIC }
/*! Select dynamic storage for component state instance*/
#define SELECT_DYNAMIC(component)\
	enum component##_storage_type { STORAGE_TYPE_TAG(component) = STATE_DYNAMIC }
/*! Select the component's initialisation type*/
#define SELECT_INITIALISATION_TYPE(component,init_type)\
	enum component##_init_type { INITIALISATION_TYPE_TAG(component) = init_type }
/*! Select 0-fill initialisation for component state instance*/
#define SELECT_ZERO_INITABLE(component)\
	SELECT_INITIALISATION_TYPE(component,ZERO_INITABLE)
/*! Select static initialisation for the component's state instance*/
#define SELECT_STATIC_INITABLE(component)\
	SELECT_INITIALISATION_TYPE(component,STATIC_INITABLE)
/*! Select user-defined initialisation for the component's state instance*/
#define SELECT_USER_INITABLE(component)\
	SELECT_INITIALISATION_TYPE(component,USER_INITABLE)
/*! Is component state in static storage ?*/
#define IS_STATIC(component)	(STORAGE_TYPE_TAG(component) == STATE_STATIC)
/*! Is component zero-initialisable?*/
#define IS_ZERO_INITABLE(component)\
	(INITIALISATION_TYPE_TAG(component) == ZERO_INITABLE)
/*! Has component a static initialiser?*/
#define HAS_STATIC_INITIALISER(component)\
	(INITIALISATION_TYPE_TAG(component) == STATIC_INITABLE)
/*! Has component user-defined initialisation and finalisation?*/
#define HAS_USER_INIT(component) (INITIALISATION_TYPE_TAG(component) == USER_INITABLE)
/*! Declare handles to component's global and public state
    when state is in dynamic storage
*/
#define DECL_DYNAMIC_STATE(component)\
	 STATE_T(component) * HANDLE(component) = NULL;\
	 PUBLIC_STATE_T(component) * PUBLIC_HANDLE(component) = NULL

/*! Declare component's state in static storage, with handles to global and
    public state.
*/
#define DECL_STATIC_STATE(component)\
	static STATE_T(component) STATE(component);\
	STATE_T(component) * HANDLE(component) = &STATE(component);\
	PUBLIC_STATE_T(component) * PUBLIC_HANDLE(component) =\
	 (PUBLIC_STATE_T(component) *)&STATE(component)

/*! Extern declaration of generated component initialising function*/
#define IMPORT_INITOR(component)	extern DECL_INITOR(component)
/*! Extern declaration of generated component finalsing function*/
#define IMPORT_FINITOR(component)	extern DECL_FINITOR(component)

/*! Extern declaration of handle to component's public state*/
#define IMPORT_STATE(component)\
extern PUBLIC_STATE_T(component) * PUBLIC_HANDLE(component)

/*! Get the value of a field from the component's global state, accessed through
    the component's handle.
*/
#define GET_STATE_BY_HANDLE(component,field)\
	(assert(true),HANDLE(component)->field)

/*! Reference a field in the component's global state, accessed through the the
    the component's handle.
*/
#define SET_STATE_BY_HANDLE(component,field) (HANDLE(component)->field)

/*! Get the value of a field from the component's global static state.
    Quicker than \c GET_STATE_BY_HANDLE for a statically implemented component.
*/
#define GET_STATIC_STATE(component,field)		(assert(true),STATE(component).field)

/*! Reference a field in the component's global static state.
    Quicker than \c SET_STATE_BY_HANDLE for a statically implemented mdoule
*/
#define SET_STATIC_STATE(component,field)		(STATE(component).field)

/*! Define the generated component initialisation function \c INITOR(component)*/
#define DEFINE_INITOR(component)\
DECL_INITOR(component) {\
	component_initor(\
		IS_STATIC(component),\
		IS_ZERO_INITABLE(component),\
		(void **)(char *)&HANDLE(component),\
		(void **)(char *)&PUBLIC_HANDLE(component),\
		STATIC_INITIALISER_HANDLE(component),\
		sizeof(STATE_T(component)),\
		USER_INIT_HANDLE(component));\
}

/*! Define the generated component finalisation function \c FINITOR(component)*/
#define DEFINE_FINITOR(component)\
DECL_FINITOR(component) {\
	component_finitor(\
		IS_STATIC(component),\
		(void **)(char *)&HANDLE(component),\
		(void **)(char *)&PUBLIC_HANDLE(component),\
		USER_INIT_HANDLE(component));\
}

/*! Initialise a component's global state
    \param is_static	Is the component's global state in static storage?
    \param is_zero_initable		Is the component's global state zero-initialisable?
    \param	state	Pointer to a pointer to the component's global state.
    \param	public_state	Pointer to a pointer to the component's public
    						state.
    \param	initialiser		\c NULL, otherwise a pointer to a constant
    						initialiser for the component's global state.
    \param	size			The size of the component's global state.
    \param	user_init		\c NULL, otherwise a pointer to a user-supplied
    						function to initialise the component's global
    						state.
    If \c is_static is false, the function asserts that <tt>*state == NULL</tt>
    and then assigns to \c *state a heap block of \c size bytes.
 *
    Otherwise, \c state is assumed to address the component's global state in
    static storage.
 *
    Then, if \c is_zero_initable is true, the component's state is 0-filled.
 *
    Otherwise, if <tt>initialiser != NULL</tt>, the \c size bytes at
    \c initialiser are copied to \c *state.
 *
    Otherwise, the function asserts that <tt>user_init != NULL</tt>, and
    then calls \c user_init(*state)
*/
extern void
component_initor(
    bool is_static,
    bool is_zero_initable,
    void **state,
    void **public_state,
    void const *initialiser,
    size_t size,
    void (*user_init)());

/*! Finalise a component's global state
    \param is_static	Is the component's global state in static storage?
    \param	state	Pointer to a pointer to the component's global state.
    \param	public_state	Pointer to a pointer to the component's public
    						state.
    \param	user_finis		\c NULL, otherwise a pointer to a user-supplied
    						function to finalise the component's global
    						state.
    If <tt>user_init != NULL</tt>, the function asserts that
    <tt>*state != NULL</tt>	and then calls \c user_init(*state)
 *
    Otherwise, \c state is assumed to address the component's global state in
    static storage.
 *
    Then, if \c is_static is false, the function again asserts that
    <tt>*state != NULL</tt>, then calls \c free(*state), and sets
    <tt>*public_state = *state = NULL</tt>
 *
*/
extern void
component_finitor(
    bool is_static,
    void **state,
    void **public_state,
    void (*user_finis)());

/*@}*/

/*! \addtogroup state_utils_user_macros

	<h3>The State Utils macros</h3>

	Macros are provided to define a component's global state structure, and
	optionally create a public state structure as part of the global
	state, for exporting to the component's clients.

	Macros are provided to create the component's state structure
	statically or dynamically (on the heap). Options are provided to specify
	that the structure is zero-initialisable, or has a constant static
	initialiser, or is to be initialised and finalised by
	user-defined functions.

	<b>Using the macros:</b>

	<ul>
		<li>
			<p><b>In my_component.h:</b></p>
			<ul>
				<li>
					<p>Define the component's public state:</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
<tt><pre>
PUBLIC_STATE_DEF(my_component) {
	int field1;
	int field2;
	...
} PUBLIC_STATE_T(my_component);</pre></tt>
						</td></tr>
					</table>
				</li>
				<br>
				<li>
					<p>Import the component's extern declarations:</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>IMPORT(my_component);</tt>
						</td></tr>
					</table>
				</li>
				<li>
					<p>Or, where the component has no state:</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt><pre>
IMPORT_INITOR(my_component);
IMPORT_FINITOR(my_component);</pre></tt>
						</td></tr>
					</table>
				</li>
			</ul>
		</li>
		<br>
		<li>
			<p><b>In my_component.c:</b></p>
			<ul>
				<li>
					<p>Define the component's gobal state:</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
<tt><pre>
STATE_DEF(my_component) {
	INCLUDE_PUBLIC(my_component); // Must be first
	int field3;
	int field4;
	...
STATE_T(my_component);</pre></tt>
						</td></tr>
					</table>
				</li>
				<br>
				<li>
					<p>
						If the component has no public state then the
						\c STATE_DEF will instead be:
					</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
<tt><pre>
STATE_DEF(my_component) {
	int field1;
	int field2;
	...
STATE_T(my_component);
NO_PUBLIC_STATE(my_component);</pre></tt>
						</td></tr>
					</table>
				</li>
				<br>
				<li>
					<p>
						If the component has no private state then instead of
						the \c STATE_DEF you need:
					</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>NO_PRIVATE_STATE(my_component);</tt>
						</td></tr>
					</table>
				</li>
				<br>
				<li>
					<p>
						Define my_component's state as static and
						zero-initialisable:
					</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>
								IMPLEMENT_STATIC(my_component,ZERO_INITABLE);
							</tt>
						</td></tr>
					</table>
				</li>
				<br>
				<li>
					<p>Or static with constant initialiser:</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
<tt><pre>
IMPLEMENT_STATIC(my_component,STATIC_INITABLE);
USE_STATIC_INITIALISER(my_component) = { { 1, 2,...}, 3, 4... };</pre></tt>
						</td></tr>
					</table>
				</li>
				<br>
				<li>
					<p>Or static with initialising and finalising functions:</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
<tt><pre>
IMPLEMENT_STATIC(my_component,USER_INITABLE);

DEFINE_USER_INIT(my_component)(STATE_T(my_mdoule) * statep)
{
	...
}

DEFINE_USER_FINIS(my_component)(STATE_T(my_component) * statep)
{
	...
}</pre></tt>
						</td></tr>
					</table>
				</li>
				<br>
				<li>
					<p>
						If you implement \c my_component with the
						\c USER_INITABLE option, you must define functions to
						intialise and finalise the state structure with
						\c DEFINE_USER_INIT and \c DEFINE_USER_FINIS.
					</p>
				</li>
				<br>
				<li>
					<p>
						Instead of \c IMPLEMENT_STATIC you may use
						\c IMPLEMENT_DYNAMIC to have the component's state structure
						allocated on the heap.
					</p>
				</li>
				<br>
				<li>
					<p>
						Alternatively you can just use \c IMPLEMENT.
						If you define \c DEFAULT_STATIC_STATE to the compiler,
						\c IMPLEMENT is short for \c IMPLEMENT_STATIC; otherwise
						it is short for \c IMPLEMENT_DYNAMIC.
					</p>
				</li>
			</ul>
		<li>
			<p><b>Initialising and cleaning up component state:</b></p>
			<ul>
				<li>
					<p>
						Before you can safely access any fields in the component's
						state structure you must call:
					</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>INITIALISE(my_component);</tt>
						</td></tr>
					</table>
				</li>
				<li>
					<p>
						If you every need to dispose of the state structure you
						must call:
					</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>FINALISE(my_component);</tt>
						</td></tr>
					</table>
				</li>
				<li>
					<p>You can reinitialise the state structure by calling:</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>REINITIALISE(my_component);</tt>
						</td></tr>
					</table>
				</li>
			</ul>
		</li>
		<br>
		<li>
			<p><b>Accessing component state:</b></p>
			<ul>
				<li>
					<p>
						Wherever the component's public state is visible, a field
						can be referenced like this:
					</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>val = GET_PUBLIC(my_component,field);</tt>
						</td></tr>
					</table>
					<br>
					<p>and can be set like this:</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>SET_PUBLIC(my_component,field) = val;</tt>
						</td></tr>
					</table>
				</li>
				<br>
				<li>
					<p>
						Wherever the component's global state is visible, a field
						can be referenced like this:
					</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>val = GET_STATE(my_component,field);</tt>
						</td></tr>
					</table>
					<br>
					<p>and set like this:</p>
					<table border="0" bgcolor="WhiteSmoke">
						<tr><td>
							<tt>SET_STATE(my_component,field) = val;</tt>
						</td></tr>
					</table>
				</li>
			</ul>
		</li>
	</ul>

	<b>Generated symbols (for debugging):</b>

	Symbols that can be generated by the
	macros for a component \c my_component are:

	- <tt><b>struct my_component_state_s</b></tt>
		-  <tt><b>my_component_state</b></tt>
		- The component's global state structure.
		- \b Note: This symbol will exist if the component is implemented
			statically.

	- <tt><b>typedef struct my_component_state_s</b></tt>
		- <tt><b>my_component_state_t</b></tt>

	- <tt><b>struct my_component_public_state_s</b></tt>
		- <tt><b>my_component_public_state</b></tt>
		- The component's public state structure.
		- \b Note: This symbol will exist if the component is implemented statically.

	- <tt><b>typedef struct my_component_public_state_s</b></tt>
		- <tt><b>my_component_public_state_t</b></tt>

	- <tt><b>struct my_component_state_s *</b></tt>
		- <tt><b>my_component_h</b></tt>
		- Pointer to the component's global state.
		- \b Note: This symbol will \em always address the component's global
			state, if it has any.

	- <tt><b>struct my_component_public_state_s *</b></tt>
		- <tt><b>my_component_public_h</b></tt>
		- Pointer to the component's public state.
		- \b Note: This symbol will \em always address the component's public
			state, if it has any.

	- <tt><b>void my_component_init(my_component_state_s *statep)</b></tt>
		- Mandatory name of user-supplied component initialising function, if any.
		You must define this function if you specify the component with the
		\c \b USER_INITABLE attribute. You can use the macro \c \b
		DEFINE_USER_INIT(my_component)	to compose the mandatory signature
		for this function.

	- <tt><b>void my_component_finis(my_component_state_s *statep)</b></tt>
		- Mandatory name of user-supplied component finalising function, if any.
		You must define this function if you specify the component with the
		\c \b USER_INITABLE attribute. You can use the macro \c \b
		DEFINE_USER_FINIS(my_component) to compose the mandatory signature
		for this function.
		- \b Note: \c my_component_init() and \c my_component_finis() do not need to
			control storage for the component's state structure. They can
			the assume the argument \c statep addresses that structure and need only
			initialise and finalise the elements of the structure.

	- <tt><b>void (*my_component_init_h)(my_component_state_s *)</b></tt>
		- If you specify the component with the \c \b USER_INITABLE attribute,
		this function pointer will be set to the address of \c my_component_init()
		and	will otherwise be \c NULL.

	- <tt><b>void (*my_component_finis_h)(my_component_state_s *)</b></tt>
		- If you specify the component with \c \b USER_INITABLE, this function
		pointer will be set to the address of \c my_component_finis()
		and	will otherwise be \c NULL.

	- <tt><b>const struct my_component_state_s</b></tt>
		- <tt><b>my_component_static_initialiser</b></tt>
		- Mandatory name of the component's constant static initialiser,
		if any. You can use the macro \c USE_STATIC_INITIALISER(my_component)
		to compose the mandatory declaration.

	- <tt><b>const struct my_component_state_s *</b></tt>
		- <tt><b>const my_component_static_initialiser_h</b></tt>
		- If you specify the component  with \c \b HAS_STATIC_INITIALISER, this
		pointer will be set to the address of \c my_component_static_initialiser
		and	will otherwise be \c NULL.

*/


/*! \addtogroup state_utils_impl_macros*/
/*@{*/

/*! The storage type of the component's state*/
enum component_storage_type {
    STATE_STATIC,
    /*!< Module's state is in static storage*/
    STATE_DYNAMIC
    /*!< Module's state is in dynamic storage*/
};

/*! The initialisation type of the component's state*/
enum component_initialisation_type {
    ZERO_INITABLE,
    /*!< Module is zero-initialisable*/
    STATIC_INITABLE,
    /*!< Module has a constant static initialiser*/
    USER_INITABLE
    /*!< Module has user-defined initialisation	and finalisation
    	functions*/
};


/*! Implement the component's state in static storage
    \param	component	Name of the component
    \param init_type	Type of component initialisation. One of:
    	- \c ZERO_INITABLE
    	- \c STATIC_INITABLE
    	- \c USER_INITABLE
*/
#define IMPLEMENT_STATIC(component,init_type)\
	DECL_STATIC_STATE(component);\
	SELECT_STATIC(component);\
	SELECT_INITIALISATION_TYPE(component,init_type);\
	DECL_STATIC_INITIALISER_HANDLE(component);\
	DECL_USER_INIT_HANDLE(component);\
	DECL_USER_FINIS_HANDLE(component);\
	DEFINE_INITOR(component)\
	DEFINE_FINITOR(component)

/*! Implement the component's state in dynamic storage
    \param	component	Name of the component
    \param init_type	Type of component initialisation. One of:
    	- \c ZERO_INITABLE
    	- \c STATIC_INITABLE
    	- \c USER_INITABLE
*/
#define IMPLEMENT_DYNAMIC(component,init_type)\
	DECL_DYNAMIC_STATE(component);\
	SELECT_DYNAMIC(component);\
	SELECT_INITIALISATION_TYPE(component,init_type);\
	DECL_STATIC_INITIALISER_HANDLE(component);\
	DECL_USER_INIT_HANDLE(component);\
	DECL_USER_FINIS_HANDLE(component);\
	DEFINE_INITOR(component)\
	DEFINE_FINITOR(component)
/* Implement the component's state dynamically*/

#ifdef DEFAULT_STATIC_STATE
/*! Abbreviate \c IMPLEMENT_STATIC as \c IMPLEMENT when \c DEFAULT_STATIC_STATE
    is defined.
*/
#define IMPLEMENT(component,init_type)	IMPLEMENT_STATIC(component,init_type)
#else
/*! Abbreviate \c IMPLEMENT_DYNAMIC as \c IMPLEMENT when \c DEFAULT_STATIC_STATE
    is not defined.
*/
#define	IMPLEMENT(component,init_type)	IMPLEMENT_DYNAMIC(component,init_type)
#endif

/*! Declare a static initialiser for the component.
    \param	 component	The name of the component.
    The macro invocation should be immediately followed by the
    specification of the static initialiser:
 *
    -	<tt>USE_STATIC_INITIALISER(component) = {...};</tt>
 *
    The static initialiser will be copied into the component's global
    state instance whenever \c INITIALISE(component) is invoked.
*/
#define USE_STATIC_INITIALISER(component)\
	DECL_STATIC_INITIALISER(component);\
	DECL_STATIC_INITIALISER_HANDLE(component) = &STATIC_INITIALISER(component);\
	DECL_STATIC_INITIALISER(component)

/*! Define a user initialisation function for component.
    \param	component	Name of the component.
    The function body should be written immediately after the macro invocation:
 *
    - <tt>DEFINE_USER_INIT(component){...};</tt>
 *
    The macro generates the function signature:
 *
    - <tt>void USER_INIT(component)(STATE_T(component) *)</tt>
 *
    A user initialisation function must be defined if the component is implemented
 * 	with initialisation type \c USER_INITABLE.	The user initialisation
    function \c USER_INIT(component) should not be confused with the
    generated component initialisation function \c INITOR(component).
    \c USER_INIT(component) does not need to provide storage for the component's
    state. It can assume that storage is addressed by the
    <tt>STATE_T(component) *</tt> argument and need only initialise the
    the elements of the state structure.
*/
#define DEFINE_USER_INIT(component)\
	DECL_USER_INIT(component);\
	DECL_USER_INIT_HANDLE(component) = &USER_INIT(component);\
	void USER_INIT(component) /* ...user_init function body here*/

/*! Define a user finalisation function for component.
    \param	component	Name of the component.
    The function body should be written immediately after the macro invocation:
 *
    - <tt>DEFINE_USER_FINIS(component){...};</tt>
 *
    The macro generates the function signature:
 *
    - <tt>void USER_FINIS(component)(STATE_T(component) *)</tt>
 *
    A user finalisation function must be defined if the component is implemented
 * 	with initialisation type \c USER_INITABLE.	The user finalisation
    function \c USER_FINIS(component) should not be confused with the
    generated component finalisation function \c FINITOR(component).
    \c USER_FINIS(component) does not need to destroy storage for the component's
    state, even if the component is implemented dynamically. That dynamic
    storage will be disposed of by \c FINITOR(component). \c USER_FINIS(component)
    need only finalise the elements of the state structure.
*/
#define DEFINE_USER_FINIS(component)\
	DECL_USER_FINIS(component);\
	DECL_USER_FINIS_HANDLE(component) = &USER_FINIS(component);\
	void USER_FINIS(component) /* ... user_finis function body here*/

/*! Invoke this macro immediately after opening \c STATE_DEF(component) to place
    the public state instance at the start of the state instance. It must be
    invoked at that position if the component has public as well as private
    state.
    \param	component	Name of the component
*/
#define INCLUDE_PUBLIC(component)	PUBLIC_STATE_T(component)	PUBLIC_STATE(component)

/*! Invoke this macro in a component's implementation file to open the definition
    of the component's global state structure:
 *
    - <tt>STATE_DEF(component) {...}</tt>
 *
    \param	component	Name of the component
*/
#define STATE_DEF(component)	typedef struct component##_state_s

/*! Invoke this macro in a component's implementation file to close the definition
    of the component's global state:
 *
    - <tt>STATE_DEF(component){...} STATE_T(component);</tt>
 *
    \param	component	Name of the component
*/
#define STATE_T(component)		component##_state_t

/*! Invoke this macro in a component's implementation file in place of
    \c STATE_DEF(component) to say that the component's global state is simply its
    public state (if any ).
    \param	component	Name of the component
*/
#define NO_PRIVATE_STATE(component)\
	typedef PUBLIC_STATE_T(component) STATE_T(component)

#ifdef DEFAULT_STATIC_STATE
/*! Abbreviate \c GET_STATIC_STATE as \c GET_STATE when
    \c DEFAULT_STATIC_STATE is defined.
 *
    Invoke this macro in the component's implementation file to
    evaluate a field in the component's global state. If the
    field also belongs to the component's public state, then
    prefer \c GET_PUBLIC.
*/
#define GET_STATE(component,field)		GET_STATIC_STATE(component,field)
/*! Abbreviate \c SET_STATIC_STATE as \c SET_STATE when
    \c DEFAULT_STATE_STATIC is defined.
 *
    Invoke this macro in the component's implementation file to
    code an assignable reference to a field in the component's global state.
 *
    - <tt>SET_STATE(component,field) = value;</tt>
 *
    \param	component	Name of the component
    \param	field	Name of the field to be evaluated in the component's state.
 *
 * 	If the field also belongs to the component's public state, then
    prefer \c SET_PUBLIC.
*/
#define SET_STATE(component,field)		SET_STATIC_STATE(component,field)
#else
/*! Abbreviate \c GET_STATE_BY_HANDLE_ as \c GET_STATE when
    \c DEFAULT_STATIC_STATE is not defined.
 *
    Invoke this macro in the component's implementation file to
    evaluate a field in the component's global state.
 *
    \param	component	Name of the component
    \param	field	Name of the field to be evaluated in the component's state.
 *
    If the field also belongs to the component's public state, then
    prefer \c GET_PUBLIC.
*/
#define GET_STATE(component,field)		GET_STATE_BY_HANDLE(component,field)
/*! Abbreviate \c SET_STATE_BY_HANDLE as \c SET_STATE when
    \c DEFAULT_STATIC_STATE is not defined.
 *
    Invoke this macro in the component's implementation file to
    code an assignable reference to a field in the component's global state.
 *
    - <tt>SET_STATE(component,field) = value;</tt>

    \param	component	Name of the component
    \param	field	Name of the field to be evaluated in the component's state.
 *
 * 	If the field also belongs to the component's public state, then
    prefer \c SET_PUBLIC.
*/
#define SET_STATE(component,field)		SET_STATE_BY_HANDLE(component,field)
#endif


/*@}*/

/*! \addtogroup state_utils_hdr_macros*/
/*@{*/

/*! Invoke this macro in a component's header file to open the definition
    of the component's public state structure:
 *
    - <tt>PUBLIC_STATE_DEF(component) {...}</tt>
 *
    \param	component	Name of the component
*/
#define PUBLIC_STATE_DEF(component)	typedef struct component##_public_state_s

/*! Invoke this macro in a component's header file to close the definition
    of the component's public state:
 *
    - <tt>PUBLIC_STATE_DEF(component){...} PUBLIC_STATE_T(component);</tt>
 *
    \param	component	Name of the component
*/
#define PUBLIC_STATE_T(component)	component##_public_state_t

/*! Invoke this macro in a component's header file in place of
    \c PUBLIC_STATE_DEF(component) to say that the component's global state is simply
    its private state (if any ).
 *
    \param	component	Name of the component
*/
#define NO_PUBLIC_STATE(component)\
typedef STATE_T(component) PUBLIC_STATE_T(component)

/*! Invoke this macro in a component's header file to expose the component's public
    state and its generated initialisation and finalisation functions.
 *
    \param	component	Name of the component
*/
#define IMPORT(component)\
	IMPORT_STATE(component);\
	IMPORT_INITOR(component);\
	IMPORT_FINITOR(component)

/*@}*/

/*! \ingroup state_utils_user_macros
    \name Macros for component client code.
*/
/*@{*/

/*! Initialise the component's state - appropriately whether static or dynamic,
    zero-initialisable, static-initialisable, user-initialisable.
    \param	component	Name of the component
*/
#define INITIALISE(component)	INITOR(component)()

/*! Finalise the component's state - appropriately whether static or dynamic,
 * 	user-finalised or not.
    \param	component	Name of the component
*/
#define FINALISE(component)	INITOR(component)()

/*! Finalise and then initialise the component's state appropriately
    \param component	Name of the component
*/
#define REINITIALISE(component)\
	(void)(FINALISE(component),INITIALISE(component))

/*!	Invoke this macro in client code file to evaluate a field in the
    component's public state.
    \param	component	Name of the component
    \param	field	Name of the field to be evaluated in the component's state.
*/
#define	GET_PUBLIC(component,field)\
	(assert(true),(PUBLIC_HANDLE(component)->field))

/*!	Invoke this macro in client code file to code an assignable
    reference to a field in the component's public state:
 *
    - <tt>SET_PUBLIC(component,field) = value;</tt>
 *
    \param	component	Name of the component
    \param	field	Name of the field to be evaluated in the component's state.
*/
#define	SET_PUBLIC(component,field)	(PUBLIC_HANDLE(component)->field)


#endif

/*@}*/

/* EOF*/
