/***************************************************************************
 *   Copyright (C) 2007-2010 Mike Kinghan, imk@strudl.org            *
 *   All rights reserved.                                                  *
 *   Contributed originally by Mike Kinghan, imk@strudl.org                *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "report.h"
#include "chew.h"
#include "symbol.h"
#include "evaluator.h"
#include "memory.h"
#include "configured_symbols.h"
#include "swiss_army.h"
#include <stdio.h>

/*! \ingroup symbol_component symbol_interface symbol_internals
	\file symbol.c
	This file implements the Symbol component
*/

/*! \addtogroup symbol_internals */
/*@{*/

/*!	Initialise a symbol.
	\param	sym	The symbol to initialise. Is assumed to control
				only unitialised storage.
	\param	name The name of this symbol.
	\param	namelen The length of \e name.
*/
static void
symbol_init(symbol_h sym, char const *name, size_t namelen)
{
    eval_result_clear(sym);
    sym->sym_name = clone(name,namelen);
}

/*!	Initialise a symbol from another
	\param	dest The symbol to initialise. Is assumed to control
				only unitialised storage.
	\param	src	The symbol to be copied.

	\e dest is initialised as a copy of \e src
*/
static void
symbol_copy_init(symbol_h dest, symbol_const_h src)
{
    dest->sym_name = clone(src->sym_name,0);
    if (src->sym_def) {
        dest->sym_def = canonical_string_copy(src->sym_def);
    } else {
        dest->sym_def = src->sym_def;
    }
    if (src->most_resolved) {
        dest->most_resolved = canonical_string_copy(src->most_resolved);
    } else {
        dest->most_resolved = src->most_resolved;
    }
    dest->flags = src->flags;
    dest->value = src->value;
}

/*!	Compare a symbol with a specified name.
	\param	sym	The symbol to compare
	\param	name	The name to compare
	\param	namelen	The length of the name.
	\return < 0 if the name of \e sym compares less than \e name, 0 if they
		compare equal, otherwise > 0
*/
static int
symbol_name_compare(symbol_const_h sym, char const *name, size_t namelen)
{
    char const *sym_name;
    int comp;
    assert(sym);
    assert(name);
    sym_name = symbol_name(sym);
    comp = strncmp(sym_name,name,namelen);
    if (comp <= 0) {	/* Symbol name compares lesser.*/
        if (comp == 0) { /* Initial match*/
            if (sym_name[namelen] != '\0') {
                /* Symbol name is longer than `name'*/
                comp = 1;
            }
            /* Otherwise full match*/
        }
    }
    return comp;
}

/*!	Handler for circular symbol definition.
	\param	sym	The symbol with a circular definition.

	A warning is emitted reporting the circular definition.
	The symbol's \e most-resolved field is set to the symbol's
	name.

	A symbol \e SYM will be found to have a circular definition if
	its name occurs as an identifier in its definition or in
	the definition of any symbol that must be resolved to resolve
	\e SYM.
*/
static
void handle_circular_sym_def(symbol_h sym)
{
    char const *sym_name;
    assert(sym);
    sym_name = sym->sym_name;
    report(GRIPE_CIRCULAR_SYM_DEF,NULL,
           "Symbol \"%s\" has a circular definition",
           sym_name);
    SET_KEEP(*sym);
    assert(!sym->most_resolved);
    sym->most_resolved =
        canonical_string_new(&sym_name,0);
}

/*! A helper function for resolving defined symbols.

	\param	sym	The symbol being resolved. On entry this
		symbol is assumed to have no most-resolved form assigned.
	\param	def	On entry, \e def is a pointer to the
		address of text that must be resolved in order
		to resolve \e sym. On return the address of the
		unresolved portion of the text is stored here.

	The function is called initially when \e sym has
	a definition for which we want the most-resolved form,
	with \e def pointing to the address of (a copy of) that definition.

	The function evaluates only the longest intial segment of
	of the text referenced by \e def that constitutes a legal
	expression.

	If this expression is soluble then \e sym assumes the resolved value
	of the expression and its most-resolved form is the
	most-resolved form of the expression.

	If the parsed expresssion is not soluble but nevertheless has a
	most-resolved form then \e sym does not assume a resolved value but does
	assume the most-resolved form of the expression (e.g. the symbol is defined as
	as FOO, where FOO is neither a --define-ed or an --undef-ed symbol).

	If the parsed expression is not soluble and has no most-resolved form then
	the expression is an empty initial segment of the text, followed
	by garbage. In this case the value stored at \e def on return will
	be unchanged from the value on entry.

	The function will detect a circular definition for \e sym when
	the name of \e sym recurs in the text that must be resolved to
	resolve \e sym. In this a warning is emitted reporting that
	\e sym is circularly defined: \e sym evaluates as insoluble and
	its most resolved form is deemed to be its own name.
*/
static void
symbol_resolve_helper_minor(symbol_h sym, char const **def)
{
    char const *cp = chew_on(*def);
    char const *end;
    /* Is the definition a numeral? */
    int_spec_t val = eval_numeral(cp,&end);
    if (*end != '\0') {
        /* Definition is not a numeral. Is it a configured symbol? */
        bool try_expr = true;	/* Assume definition may be expression */
        val.type = INT_INSOLUBLE; /* Assume we won't be able to resolve */
        if (end == cp && is_symbol_start_char(*cp)) {
            /* No initial numeral. Possible symbol. */
            symbol_h next_sym = configured_symbol_seek(&end);
            if (next_sym && *end == '\0') {
                /* Definition is a configured symbol */
                try_expr = false;	/* So don't bother trying expresssion */
                if (next_sym == sym) {
                    /* Sym has circular definition! */
                    handle_circular_sym_def(sym);
                } else if (symbol_def_len(next_sym) < 1) {
                    /* next_sym has no definition. */
                    symbol_resolve(next_sym);
                    if (RESOLVED(*next_sym)) {
                        val = next_sym->value;
                    } else {
                        SET_KEEP(*sym);
                    }
                    /* The most resolved form of sym is the same as that
                        of next_sym */
                    sym->most_resolved =
                        canonical_string_copy(next_sym->most_resolved);
                } else {
                    /* next_sym has definition. Recurse on the definition */
                    char const * next_sym_def = cp =
                                                    clone(canonical_string_text(next_sym->sym_def),0);
                    symbol_resolve_helper_minor(sym,&cp);
                    free((void *)next_sym_def);
                    if (!sym->most_resolved) {
                        sym->most_resolved =
                            canonical_string_copy(next_sym->sym_def);
                    }
                }
            } else if (*end == '\0') {
                /* Definition is an unconfigured symbol. */
                try_expr = false;	/* But don't try expression */
                assert(!sym->most_resolved);
                /*! Most resolved form is definition itself */
                sym->most_resolved = canonical_string_new(&cp,0);
            }
        }
        if (try_expr) {
            /* Definition is not a numeral or a configured
                    or unknown symbol.	Is it an expression? */
            eval_result_t er = EVAL_RESULT_INITOR;
            end = cp;
            er = eval_definition_text(&end);
            if (RESOLVED(er)) {
                /* Definition begins with a soluble expression */
                val = er.value;
            } else if (er.sym_name) {
                /* Definition begins with a configured symbol that
                    isn't soluble */
                if (!strcmp(er.sym_name,sym->sym_name)) {
                    /* sym has a circular definition! */
                    handle_circular_sym_def(sym);
                } else {
                    SET_KEEP(*sym);
                    sym->most_resolved =
                        canonical_string_copy(er.most_resolved);
                }
            } else {
                /* Definition begins with indeterminate stuff */
                SET_KEEP(*sym);
                end = cp;	/* Consume none */
            }
        }
    }
    if (val.type != INT_INSOLUBLE) { /* Got a resolved value at some point */
        canonical_string_h num = canonical_numeral(&val);
        eval_result_set_value(sym,&val);
        sym->most_resolved = num;
    }
    *def = end; /* Store start of unconsumed text */
}

/*! A helper function for resolving defined symbols.

	\param	sym	The symbol being resolved. On entry this
		symbol is assumed to have no most-resolved form assigned.
	\param	def	A pointer to the 0-terminated definition of \e sym.

	The function parses the whole of \e def and assembles from \e def the
	most-resolved form that it assigns to \e sym.

	If \e def constitutes a legal expression then the call equates to
	single call to \e symbol_resolve_helper_minor(sym,&def). On return
	from that call, \e def will address an empty string and the
	function will return.

	Otherwise the function will initialise the most-resolved form that will
	finally be assigned to \e sym with any most-resolved form that has
	so far accrued to \e sym, and then consume the remainder of \e def. as
	sequence of segments, separated by whitespace and/or comments, each
	of which is either a legal expression or garbage. Iterated calls to
	\e symbol_resolve_helper_minor(sym,&cp), are made, where \e cp points to the
	the unconsumed remainder of \e def at each call. When this call
	consumes any of the remaining text, the most-resolved form that has accrued

	to \e sym by the call is appended to the most-resolved form under
	construction and then deleted from \e sym. When the call consumes none
	of the remaining text, a token of garbage is consumed from the start
	of the remaining text and appended to the most-resolved form under
	construction.

	When \e def is completely consumed in this way, the most-resolved form
	that has been assembled from it is assigned to \e sym, and \e sym
	is flagged as a not having a legal value.
*/

static void
symbol_resolve_helper_major(symbol_h sym, char const *def)
{
    char const *cp = def;
    char const *end = cp;
    canonical_string_h lparen = NULL;
    canonical_string_h rparen = NULL;
    /* If the definition is a legal expression this call will
    	consume all of it */
    SET_PUBLIC(chew,in_source) = false;
    symbol_resolve_helper_minor(sym,&end);
    SET_PUBLIC(chew,in_source) = true;
    SET_PUBLIC(chew,in_double_quote) = false; /* Forget any unclosed quotation in symbol definition */
    if (*end) {
        /* Definition is not a legal expression, but
        		we may have resolved an expression in an initial part */
        /* Init canonical string to build most-resolved form */
        canonical_string_h most_resolved = canonical_string_new(NULL,0);
        for (	; *cp; SET_UNRESOLVED(*sym)) {
            if (cp != end && *cp == '(') {
                /*  Since cp != end, parsing has stopped where legal syntax breaks.
                	So we can have parentheses marooned to the left of
                	the expression we have resolved. Append all these
                	to the most-resolved form of def, before resolved expression. */
                do {
                    if (!lparen) {
                        char const * parenl = "(";
                        /* Build a canonical string from "(" just once */
                        lparen = canonical_string_new(&parenl,0);
                    }
                    canonical_string_append(most_resolved,lparen,true);
                } while (*++cp == '(');
                cp = end = chew_on(end);
            }
            if (sym->most_resolved) { /* We have resolved part of definition */
                /* Append its most resolved form to what we have got */
                canonical_string_append(most_resolved,sym->most_resolved,true);
                canonical_string_dispose(sym->most_resolved);
                /* Clear down most-resolved appendage for next iteration */
                sym->most_resolved = NULL;
                cp = end = chew_on(end);
            }
            /* Have not resolved any more of the definition this time around */
            else if (*cp == ')') {
                /* Found a right-hand paren cut off from
                		its match by syntax break */
                do { /* Append all these to the most-resolved form of def */
                    if (!rparen) {
                        char const * parenr = ")";
                        /* Build a canonical string from ")" just once */
                        rparen = canonical_string_new(&parenr,0);
                    }
                    canonical_string_append(most_resolved,rparen,true);
                } while(*++cp == ')');
                end = cp = chew_on(cp);
            } else { /* Something unresolved other than ")" */
                if (end == cp) {
                    /* Evaluation didn't consume any text
                    		so consume next string */
                    end = chew_string(cp);
                }
                if (end > cp) {
                    /* Have consumed some token. Append it to most-resolved
                    	form of def */
                    heap_str token = clone(cp,end - cp);
                    char const * tp = token;
                    canonical_string_h cs = canonical_string_new(&tp,0);
                    canonical_string_append(most_resolved,cs,true);
                    canonical_string_dispose(cs);
                    free(token);
                }
                cp = end = chew_on(end);
            }
            /* Now try to resolve some more of the definition */
            symbol_resolve_helper_minor(sym,&end);
            if (*cp != '(') {
                cp = end;
            }
        }
        canonical_string_dispose(lparen);
        canonical_string_dispose(rparen);
        SET_ILLEGAL(*sym);
        assert(!sym->most_resolved);
        sym->most_resolved = most_resolved;
    }
}

/*@}*/

/* API*/

symbol_h
symbol_new(char const *name, size_t namelen)
{
    symbol_h sym;
    assert(name);
    assert(namelen);
    sym = zallocate(sizeof(eval_result_t));
    symbol_init(sym,name,namelen);
    return sym;
}

void
symbol_dispose(symbol_h sym)
{
    if (sym) {
        eval_result_clear(sym);
        free(sym);
    }
}

symbol_h
symbol_copy(symbol_const_h src)
{
    symbol_h dest;
    assert(src);
    dest = zallocate(sizeof(eval_result_t));
    symbol_copy_init(dest,src);
    return dest;
}

void
symbol_swap(symbol_h lhs, symbol_h rhs)
{
    assert(lhs);
    assert(rhs);
    PODSWAP(char *,&lhs->sym_name,&rhs->sym_name);
    canonical_string_swap(lhs->sym_def,rhs->sym_def);
    canonical_string_swap(lhs->most_resolved,rhs->most_resolved);
    PODSWAP(int_spec_t,&lhs->value,&lhs->value);
    PODSWAP(int,&lhs->flags,&lhs->flags);
}

void
symbol_assign(symbol_h dest, symbol_const_h src)
{
    if (dest != src) {
        symbol_h tmp = symbol_copy(src);
        symbol_swap(dest,tmp);
        symbol_dispose(tmp);
    }
}

bool
symbol_equal(symbol_const_h lhs,symbol_const_h rhs)
{
    bool eq = lhs == rhs;
    if (!eq) {
        eq = symbol_compare(lhs,rhs,0);
        if (eq) {
            eq = canonical_string_equal(symbol_definition(lhs),
                                        symbol_definition(rhs));
        }
    }
    return eq;
}

void
symbol_define(symbol_h sym, char const **pdef)
{
    assert(sym);
    SET_PUBLIC(chew,in_source) = false;
    canonical_string_dispose(sym->sym_def);
    sym->sym_def = canonical_string_new(pdef,0);
    canonical_string_dispose(sym->most_resolved);
    sym->most_resolved = NULL;
    SET_PUBLIC(chew,in_source) = false;
}

void
symbol_undefine(symbol_h sym)
{
    assert(sym);
    canonical_string_dispose(sym->sym_def);
    sym->sym_def = NULL;
    canonical_string_dispose(sym->most_resolved);
    sym->most_resolved = NULL;
    eval_result_set_value(sym,0);
}

char const *
symbol_name(symbol_const_h sym)
{
    assert(sym);
    return sym->sym_name;
}

canonical_string_const_h
symbol_definition(symbol_const_h sym)
{
    assert(sym);
    return sym->sym_def;
}

canonical_string_const_h
symbol_final_value(symbol_h sym)
{
    assert(sym);
    symbol_resolve(sym);
    return sym->sym_def ? sym->most_resolved : NULL;
}

int
symbol_def_len(symbol_const_h sym)
{
    assert(sym);
    return sym->sym_def ? canonical_string_length(sym->sym_def) : -1;
}

void
symbol_memorise(symbol_h sym)
{
    assert(sym);
    sym->flags |= EVAL_SEEN;
}

bool
symbol_seen(symbol_const_h sym)
{
    assert(sym);
    return (sym->flags & EVAL_SEEN) != 0;
}

void
symbol_forget(symbol_h sym)
{
    assert(sym);
    sym->flags &= ~EVAL_SEEN;
}

int
symbol_compare(symbol_const_h lhs, void const * rhs, size_t len)
{
    assert(lhs);
    assert(rhs);
    if (len) {
        return symbol_name_compare(lhs,rhs,len);
    } else {
        symbol_const_h other = rhs;
        return strcmp(lhs->sym_name,other->sym_name);
    }
}

bool
symbol_is_legal(char const *tok, size_t toklen)
{
    if (!is_symbol_start_char(*tok) || toklen < 1) {
        return false;
    }
    for (--toklen; toklen; --toklen,++tok) {
        if (!is_symbol_inner_char(*tok)) {
            return false;
        }
    }
    return true;
}

bool
is_symbol_start_char(char cp)
{
    return isalpha(cp) || cp == '_';
}

bool
is_symbol_inner_char(char cp)
{
    return isalnum(cp) || cp == '_';
}

void
symbol_resolve(symbol_h sym)
{
    if (!symbol_seen(sym)) {
        size_t def_len = symbol_def_len(sym);
        symbol_memorise(sym);
        switch(def_len) {
        case -1:	/* --undef SYM */
            eval_result_set_value(sym,&int_spec_false);
            sym->most_resolved = canonical_numeral(&int_spec_false);
            break;
        case 0:		/* --define SYM */
            SET_KEEP(*sym);
            sym->most_resolved = canonical_string_new(NULL,0);
            break;
        default: {
            heap_str sym_def =
                clone(canonical_string_text(sym->sym_def),0);
            symbol_resolve_helper_major(sym,sym_def);
            free(sym_def);
        }
        }
    }
}



/* EOF*/
